# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An OrderedSet is a custom MutableSet that remembers its order, so that every
entry has an index that can be looked up. It can also act like a Sequence.

Based on a recipe originally posted to ActiveState Recipes by Raymond Hettiger,
and released under the MIT license.
"""

import itertools as it
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    MutableSet,
    AbstractSet,
    Sequence,
    Set,
    TypeVar,
    Union,
    overload,
)

SLICE_ALL = slice(None)
__version__ = "4.1.0"


T = TypeVar("T")

# SetLike[T] is either a set of elements of type T, or a sequence, which
# we will convert to an OrderedSet by adding its elements in order.
SetLike = Union[AbstractSet[T], Sequence[T], Iterable[T]]
OrderedSetInitializer = Union[AbstractSet[T], Sequence[T], Iterable[T]]


def _is_atomic(obj: Any) -> bool:
    """
    Returns True for objects which are iterable but should not be iterated in
    the context of indexing an OrderedSet.

    When we index by an iterable, usually that means we're being asked to look
    up a list of things.

    However, in the case of the .index() method, we shouldn't handle strings
    and tuples like other iterables. They're not sequences of things to look
    up, they're the single, atomic thing we're trying to find.

    As an example, oset.index('hello') should give the index of 'hello' in an
    OrderedSet of strings. It shouldn't give the indexes of each individual
    character.
    """
    return isinstance(obj, str) or isinstance(obj, tuple)


class OrderedSet(MutableSet[T], Sequence[T]):
    """
    An OrderedSet is a custom MutableSet that remembers its order, so that
    every entry has an index that can be looked up.

    Example:
        >>> OrderedSet([1, 1, 2, 3, 2])
        OrderedSet([1, 2, 3])
    """

    def __init__(self, initial: OrderedSetInitializer[T] = None):
        self.items: List[T] = []
        self.map: Dict[T, int] = {}
        if initial is not None:
            # In terms of duck-typing, the default __ior__ is compatible with
            # the types we use, but it doesn't expect all the types we
            # support as values for `initial`.
            self |= initial  # type: ignore

    def __len__(self):
        """
        Returns the number of unique elements in the ordered set

        Example:
            >>> len(OrderedSet([]))
            0
            >>> len(OrderedSet([1, 2]))
            2
        """
        return len(self.items)

    @overload
    def __getitem__(self, index: slice) -> "OrderedSet[T]":
        ...

    @overload
    def __getitem__(self, index: Sequence[int]) -> List[T]:
        ...

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    # concrete implementation
    def __getitem__(self, index):
        """
        Get the item at a given index.

        If `index` is a slice, you will get back that slice of items, as a
        new OrderedSet.

        If `index` is a list or a similar iterable, you'll get a list of
        items corresponding to those indices. This is similar to NumPy's
        "fancy indexing". The result is not an OrderedSet because you may ask
        for duplicate indices, and the number of elements returned should be
        the number of elements asked for.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset[1]
            2
        """
        if isinstance(index, slice) and index == SLICE_ALL:
            return self.copy()
        elif isinstance(index, Iterable):
            return [self.items[i] for i in index]
        elif isinstance(index, slice) or hasattr(index, "__index__"):
            result = self.items[index]
            if isinstance(result, list):
                return self.__class__(result)
            else:
                return result
        else:
            raise TypeError("Don't know how to index an OrderedSet by %r" % index)

    def copy(self) -> "OrderedSet[T]":
        """
        Return a shallow copy of this object.

        Example:
            >>> this = OrderedSet([1, 2, 3])
            >>> other = this.copy()
            >>> this == other
            True
            >>> this is other
            False
        """
        return self.__class__(self)

    # Define the gritty details of how an OrderedSet is serialized as a pickle.
    # We leave off type annotations, because the only code that should interact
    # with these is a generalized tool such as pickle.
    def __getstate__(self):
        if len(self) == 0:
            # In pickle, the state can't be an empty list.
            # We need to return a truthy value, or else __setstate__ won't be run.
            #
            # This could have been done more gracefully by always putting the state
            # in a tuple, but this way is backwards- and forwards- compatible with
            # previous versions of OrderedSet.
            return (None,)
        else:
            return list(self)

    def __setstate__(self, state):
        if state == (None,):
            self.__init__([])
        else:
            self.__init__(state)

    def __contains__(self, key: Any) -> bool:
        """
        Test if the item is in this ordered set.

        Example:
            >>> 1 in OrderedSet([1, 3, 2])
            True
            >>> 5 in OrderedSet([1, 3, 2])
            False
        """
        return key in self.map

    # Technically type-incompatible with MutableSet, because we return an
    # int instead of nothing. This is also one of the things that makes
    # OrderedSet convenient to use.
    def add(self, key: T) -> int:
        """
        Add `key` as an item to this OrderedSet, then return its index.

        If `key` is already in the OrderedSet, return the index it already
        had.

        Example:
            >>> oset = OrderedSet()
            >>> oset.append(3)
            0
            >>> print(oset)
            OrderedSet([3])
        """
        if key not in self.map:
            self.map[key] = len(self.items)
            self.items.append(key)
        return self.map[key]

    append = add

    def update(self, sequence: SetLike[T]) -> int:
        """
        Update the set with the given iterable sequence, then return the index
        of the last element inserted.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset.update([3, 1, 5, 1, 4])
            4
            >>> print(oset)
            OrderedSet([1, 2, 3, 5, 4])
        """
        item_index = 0
        try:
            for item in sequence:
                item_index = self.add(item)
        except TypeError:
            raise ValueError(
                "Argument needs to be an iterable, got %s" % type(sequence)
            )
        return item_index

    @overload
    def index(self, key: Sequence[T]) -> List[int]:
        ...

    @overload
    def index(self, key: T) -> int:
        ...

    # concrete implementation
    def index(self, key):
        """
        Get the index of a given entry, raising an IndexError if it's not
        present.

        `key` can be an iterable of entries that is not a string, in which case
        this returns a list of indices.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset.index(2)
            1
        """
        if isinstance(key, Iterable) and not _is_atomic(key):
            return [self.index(subkey) for subkey in key]
        return self.map[key]

    # Provide some compatibility with pd.Index
    get_loc = index
    get_indexer = index

    def pop(self, index=-1) -> T:
        """
        Remove and return item at index (default last).

        Raises KeyError if the set is empty.
        Raises IndexError if index is out of range.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset.pop()
            3
        """
        if not self.items:
            raise KeyError("Set is empty")

        elem = self.items[index]
        del self.items[index]
        del self.map[elem]
        return elem

    def discard(self, key: T) -> None:
        """
        Remove an element.  Do not raise an exception if absent.

        The MutableSet mixin uses this to implement the .remove() method, which
        *does* raise an error when asked to remove a non-existent item.

        Example:
            >>> oset = OrderedSet([1, 2, 3])
            >>> oset.discard(2)
            >>> print(oset)
            OrderedSet([1, 3])
            >>> oset.discard(2)
            >>> print(oset)
            OrderedSet([1, 3])
        """
        if key in self:
            i = self.map[key]
            del self.items[i]
            del self.map[key]
            for k, v in self.map.items():
                if v >= i:
                    self.map[k] = v - 1

    def clear(self) -> None:
        """
        Remove all items from this OrderedSet.
        """
        del self.items[:]
        self.map.clear()

    def __iter__(self) -> Iterator[T]:
        """
        Example:
            >>> list(iter(OrderedSet([1, 2, 3])))
            [1, 2, 3]
        """
        return iter(self.items)

    def __reversed__(self) -> Iterator[T]:
        """
        Example:
            >>> list(reversed(OrderedSet([1, 2, 3])))
            [3, 2, 1]
        """
        return reversed(self.items)

    def __repr__(self) -> str:
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self))

    def __eq__(self, other: Any) -> bool:
        """
        Returns true if the containers have the same items. If `other` is a
        Sequence, then order is checked, otherwise it is ignored.

        Example:
            >>> oset = OrderedSet([1, 3, 2])
            >>> oset == [1, 3, 2]
            True
            >>> oset == [1, 2, 3]
            False
            >>> oset == [2, 3]
            False
            >>> oset == OrderedSet([3, 2, 1])
            False
        """
        if isinstance(other, Sequence):
            # Check that this OrderedSet contains the same elements, in the
            # same order, as the other object.
            return list(self) == list(other)
        try:
            other_as_set = set(other)
        except TypeError:
            # If `other` can't be converted into a set, it's not equal.
            return False
        else:
            return set(self) == other_as_set

    def union(self, *sets: SetLike[T]) -> "OrderedSet[T]":
        """
        Combines all unique items.
        Each items order is defined by its first appearance.

        Example:
            >>> oset = OrderedSet.union(OrderedSet([3, 1, 4, 1, 5]), [1, 3], [2, 0])
            >>> print(oset)
            OrderedSet([3, 1, 4, 5, 2, 0])
            >>> oset.union([8, 9])
            OrderedSet([3, 1, 4, 5, 2, 0, 8, 9])
            >>> oset | {10}
            OrderedSet([3, 1, 4, 5, 2, 0, 10])
        """
        cls: type = OrderedSet
        if isinstance(self, OrderedSet):
            cls = self.__class__
        containers = map(list, it.chain([self], sets))
        items = it.chain.from_iterable(containers)
        return cls(items)

    def __and__(self, other: SetLike[T]) -> "OrderedSet[T]":
        # the parent implementation of this is backwards
        return self.intersection(other)

    def intersection(self, *sets: SetLike[T]) -> "OrderedSet[T]":
        """
        Returns elements in common between all sets. Order is defined only
        by the first set.

        Example:
            >>> oset = OrderedSet.intersection(OrderedSet([0, 1, 2, 3]), [1, 2, 3])
            >>> print(oset)
            OrderedSet([1, 2, 3])
            >>> oset.intersection([2, 4, 5], [1, 2, 3, 4])
            OrderedSet([2])
            >>> oset.intersection()
            OrderedSet([1, 2, 3])
        """
        cls: type = OrderedSet
        items: OrderedSetInitializer[T] = self
        if isinstance(self, OrderedSet):
            cls = self.__class__
        if sets:
            common = set.intersection(*map(set, sets))
            items = (item for item in self if item in common)
        return cls(items)

    def difference(self, *sets: SetLike[T]) -> "OrderedSet[T]":
        """
        Returns all elements that are in this set but not the others.

        Example:
            >>> OrderedSet([1, 2, 3]).difference(OrderedSet([2]))
            OrderedSet([1, 3])
            >>> OrderedSet([1, 2, 3]).difference(OrderedSet([2]), OrderedSet([3]))
            OrderedSet([1])
            >>> OrderedSet([1, 2, 3]) - OrderedSet([2])
            OrderedSet([1, 3])
            >>> OrderedSet([1, 2, 3]).difference()
            OrderedSet([1, 2, 3])
        """
        cls = self.__class__
        items: OrderedSetInitializer[T] = self
        if sets:
            other = set.union(*map(set, sets))
            items = (item for item in self if item not in other)
        return cls(items)

    def issubset(self, other: SetLike[T]) -> bool:
        """
        Report whether another set contains this set.

        Example:
            >>> OrderedSet([1, 2, 3]).issubset({1, 2})
            False
            >>> OrderedSet([1, 2, 3]).issubset({1, 2, 3, 4})
            True
            >>> OrderedSet([1, 2, 3]).issubset({1, 4, 3, 5})
            False
        """
        if len(self) > len(other):  # Fast check for obvious cases
            return False
        return all(item in other for item in self)

    def issuperset(self, other: SetLike[T]) -> bool:
        """
        Report whether this set contains another set.

        Example:
            >>> OrderedSet([1, 2]).issuperset([1, 2, 3])
            False
            >>> OrderedSet([1, 2, 3, 4]).issuperset({1, 2, 3})
            True
            >>> OrderedSet([1, 4, 3, 5]).issuperset({1, 2, 3})
            False
        """
        if len(self) < len(other):  # Fast check for obvious cases
            return False
        return all(item in self for item in other)

    def symmetric_difference(self, other: SetLike[T]) -> "OrderedSet[T]":
        """
        Return the symmetric difference of two OrderedSets as a new set.
        That is, the new set will contain all elements that are in exactly
        one of the sets.

        Their order will be preserved, with elements from `self` preceding
        elements from `other`.

        Example:
            >>> this = OrderedSet([1, 4, 3, 5, 7])
            >>> other = OrderedSet([9, 7, 1, 3, 2])
            >>> this.symmetric_difference(other)
            OrderedSet([4, 5, 9, 2])
        """
        cls: type = OrderedSet
        if isinstance(self, OrderedSet):
            cls = self.__class__
        diff1 = cls(self).difference(other)
        diff2 = cls(other).difference(self)
        return diff1.union(diff2)

    def _update_items(self, items: list) -> None:
        """
        Replace the 'items' list of this OrderedSet with a new one, updating
        self.map accordingly.
        """
        self.items = items
        self.map = {item: idx for (idx, item) in enumerate(items)}

    def difference_update(self, *sets: SetLike[T]) -> None:
        """
        Update this OrderedSet to remove items from one or more other sets.

        Example:
            >>> this = OrderedSet([1, 2, 3])
            >>> this.difference_update(OrderedSet([2, 4]))
            >>> print(this)
            OrderedSet([1, 3])

            >>> this = OrderedSet([1, 2, 3, 4, 5])
            >>> this.difference_update(OrderedSet([2, 4]), OrderedSet([1, 4, 6]))
            >>> print(this)
            OrderedSet([3, 5])
        """
        items_to_remove = set()  # type: Set[T]
        for other in sets:
            items_as_set = set(other)  # type: Set[T]
            items_to_remove |= items_as_set
        self._update_items([item for item in self.items if item not in items_to_remove])

    def intersection_update(self, other: SetLike[T]) -> None:
        """
        Update this OrderedSet to keep only items in another set, preserving
        their order in this set.

        Example:
            >>> this = OrderedSet([1, 4, 3, 5, 7])
            >>> other = OrderedSet([9, 7, 1, 3, 2])
            >>> this.intersection_update(other)
            >>> print(this)
            OrderedSet([1, 3, 7])
        """
        other = set(other)
        self._update_items([item for item in self.items if item in other])

    def symmetric_difference_update(self, other: SetLike[T]) -> None:
        """
        Update this OrderedSet to remove items from another set, then
        add items from the other set that were not present in this set.

        Example:
            >>> this = OrderedSet([1, 4, 3, 5, 7])
            >>> other = OrderedSet([9, 7, 1, 3, 2])
            >>> this.symmetric_difference_update(other)
            >>> print(this)
            OrderedSet([4, 5, 9, 2])
        """
        items_to_add = [item for item in other if item not in self]
        items_to_remove = set(other)
        self._update_items(
            [item for item in self.items if item not in items_to_remove] + items_to_add
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import choice
from abc import ABC, abstractmethod
from typing import Dict, Tuple, Optional

from ...utils import implements
from ..core import ActorRef
from ..errors import NoIdleSlot
from .config import ActorPoolConfig
from .message import _MessageBase


allocated_value = Tuple["AllocateStrategy", Optional[_MessageBase]]
allocated_values = Dict[Optional[ActorRef], allocated_value]
allocated_type = Dict[str, allocated_values]


class AllocateStrategy(ABC):
    __slots__ = ()

    @abstractmethod
    def get_allocated_address(
        self, config: ActorPoolConfig, allocated: allocated_type
    ) -> str:
        """
        Get external address where the actor allocated to.

        Parameters
        ----------
        config: ActorPoolConfig
            Actor pool config.
        allocated:
            Already allocated of actor and its strategy.

        Returns
        -------
        allocated_address: str
            External address to allocate.
        """


class AddressSpecified(AllocateStrategy):
    __slots__ = ("address",)

    def __init__(self, address):
        self.address = address

    @implements(AllocateStrategy.get_allocated_address)
    def get_allocated_address(
        self, config: ActorPoolConfig, allocated: allocated_type
    ) -> str:
        return self.address


class MainPool(AllocateStrategy):
    __slots__ = ()

    @implements(AllocateStrategy.get_allocated_address)
    def get_allocated_address(
        self, config: ActorPoolConfig, allocated: allocated_type
    ) -> str:
        # allocate to main process
        main_process_index = config.get_process_indexes()[0]
        return config.get_external_address(main_process_index)


class Random(AllocateStrategy):
    __slots__ = ()

    @implements(AllocateStrategy.get_allocated_address)
    def get_allocated_address(
        self, config: ActorPoolConfig, allocated: allocated_type
    ) -> str:
        return choice(config.get_external_addresses())


class RandomSubPool(AllocateStrategy):
    __slots__ = ()

    @implements(AllocateStrategy.get_allocated_address)
    def get_allocated_address(
        self, config: ActorPoolConfig, allocated: allocated_type
    ) -> str:
        return choice(config.get_external_addresses()[1:])


class ProcessIndex(AllocateStrategy):
    __slots__ = ("process_index",)

    def __init__(self, process_index: int):
        self.process_index = process_index

    @implements(AllocateStrategy.get_allocated_address)
    def get_allocated_address(
        self, config: ActorPoolConfig, allocated: allocated_type
    ) -> str:
        actual_process_index = config.get_process_indexes()[self.process_index]
        return config.get_pool_config(actual_process_index)["external_address"][0]


class RandomLabel(AllocateStrategy):
    __slots__ = ("label",)

    def __init__(self, label):
        self.label = label

    @implements(AllocateStrategy.get_allocated_address)
    def get_allocated_address(
        self, config: ActorPoolConfig, allocated: allocated_type
    ) -> str:
        return choice(config.get_external_addresses(label=self.label))


class IdleLabel(AllocateStrategy):
    __slots__ = "label", "mark"

    def __init__(self, label, mark):
        self.label = label
        self.mark = mark

    def __hash__(self):
        return hash((type(self), self.label, self.mark))

    def __eq__(self, other):
        return (
            isinstance(other, IdleLabel)
            and self.label == other.label
            and self.mark == other.mark
        )

    @implements(AllocateStrategy.get_allocated_address)
    def get_allocated_address(
        self, config: ActorPoolConfig, allocated: allocated_type
    ) -> str:
        addresses = config.get_external_addresses(label=self.label)
        for addr in addresses:
            occupied = False
            for strategy, _ in allocated.get(addr, dict()).values():
                if strategy == self:
                    occupied = True
                    break
            if not occupied:
                return addr
        raise NoIdleSlot(
            f"No idle slot for creating actor "
            f"with label {self.label}, mark {self.mark}"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import uuid
from typing import Dict, List, Optional, Tuple

from ..lib.aio import AioFilesystem
from ..lib.filesystem import FileSystem, get_fs
from ..serialization import AioSerializer, AioDeserializer
from ..utils import mod_hash, implements
from .base import StorageBackend, ObjectInfo, StorageLevel, register_storage_backend
from .core import StorageFileObject


@register_storage_backend
class FileSystemStorage(StorageBackend):
    name = "filesystem"

    def __init__(
        self, fs: FileSystem, root_dirs: List[str], level: StorageLevel, size: int
    ):
        self._fs = AioFilesystem(fs)
        self._root_dirs = root_dirs
        self._level = level
        self._size = size

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        root_dirs = kwargs.pop("root_dirs")
        level = kwargs.pop("level")
        size = kwargs.pop("size", None)
        fs = kwargs.pop("fs", None)
        if kwargs:  # pragma: no cover
            raise TypeError(
                f'FileSystemStorage got unexpected config: {",".join(kwargs)}'
            )

        if isinstance(root_dirs, str):
            root_dirs = root_dirs.split(":")
        if isinstance(level, str):
            level = StorageLevel.from_str(level)

        if fs is None:
            fs = get_fs(root_dirs[0])

        for d in root_dirs:
            if not fs.exists(d):
                fs.mkdir(d)
        params = dict(fs=fs, root_dirs=root_dirs, level=level, size=size)
        return params, params

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        fs = kwargs.get("fs")
        root_dirs = kwargs.get("root_dirs")
        for d in root_dirs:
            fs.delete(d, recursive=True)

    @property
    @implements(StorageBackend.level)
    def level(self) -> StorageLevel:
        return self._level

    @property
    @implements(StorageBackend.size)
    def size(self) -> Optional[int]:
        return self._size

    def _generate_path(self):
        file_name = str(uuid.uuid4())
        selected_index = mod_hash(file_name, len(self._root_dirs))
        selected_dir = self._root_dirs[selected_index]
        return os.path.join(selected_dir, file_name)

    @implements(StorageBackend.get)
    async def get(self, object_id, **kwargs) -> object:
        if kwargs:  # pragma: no cover
            raise NotImplementedError(f'Got unsupported args: {",".join(kwargs)}')

        file = await self._fs.open(object_id, "rb")
        async with file as f:
            deserializer = AioDeserializer(f)
            return await deserializer.run()

    @implements(StorageBackend.put)
    async def put(self, obj, importance: int = 0) -> ObjectInfo:
        serializer = AioSerializer(obj)
        buffers = await serializer.run()
        buffer_size = sum(getattr(buf, "nbytes", len(buf)) for buf in buffers)

        path = self._generate_path()
        file = await self._fs.open(path, "wb")
        async with file as f:
            for buffer in buffers:
                await f.write(buffer)

        return ObjectInfo(size=buffer_size, object_id=path)

    @implements(StorageBackend.delete)
    async def delete(self, object_id):
        await self._fs.delete(object_id)

    @implements(StorageBackend.list)
    async def list(self) -> List:
        file_list = []
        for d in self._root_dirs:
            file_list.extend(list(await self._fs.ls(d)))
        return file_list

    @implements(StorageBackend.object_info)
    async def object_info(self, object_id) -> ObjectInfo:
        stat = await self._fs.stat(object_id)
        return ObjectInfo(size=stat["size"], object_id=object_id)

    @implements(StorageBackend.open_writer)
    async def open_writer(self, size=None) -> StorageFileObject:
        path = self._generate_path()
        file = await self._fs.open(path, "wb")
        return StorageFileObject(file, file.name)

    @implements(StorageBackend.open_reader)
    async def open_reader(self, object_id) -> StorageFileObject:
        file = await self._fs.open(object_id, "rb")
        return StorageFileObject(file, file.name)


@register_storage_backend
class DiskStorage(FileSystemStorage):
    name = "disk"

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        kwargs["level"] = StorageLevel.DISK
        return await super().setup(**kwargs)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time
from typing import List

import numpy as np

from ... import opcodes as OperandDef
from ...core.context import get_context
from ...oscar import ActorNotExist
from ...serialization.serializables import (
    FieldTypes,
    KeyField,
    StringField,
    DictField,
    TupleField,
)
from ...lib.filesystem import open_file
from ...utils import has_unknown_shape
from ..datasource import tensor as astensor
from .core import TensorDataStore


class _HDF5Container:
    def __init__(self, all_chunk_op_keys: List[str]):
        self._all_chunk_op_keys = set(all_chunk_op_keys)
        self._done_chunk_op_keys = set()
        self._lock = threading.Lock()

    def acquire(self):
        return self._lock.acquire()

    def release(self):
        return self._lock.release()

    def mark_done(self, op_key: str):
        self._done_chunk_op_keys.add(op_key)

    def is_done(self):
        return self._done_chunk_op_keys == self._all_chunk_op_keys


class TensorHDF5DataStore(TensorDataStore):
    _op_type_ = OperandDef.TENSOR_STORE_HDF5

    _input = KeyField("input")
    _filename = StringField("filename")
    _group = StringField("group")
    _dataset = StringField("dataset")
    _dataset_kwds = DictField("dataset_kwds", key_type=FieldTypes.string)
    _axis_offsets = TupleField("axis_offsets", FieldTypes.int32)
    _out_shape = TupleField("out_shape", FieldTypes.int32)
    _container_name = StringField("container_name")

    def __init__(
        self,
        filename=None,
        group=None,
        dataset=None,
        dataset_kwds=None,
        container_name=None,
        **kw,
    ):
        super().__init__(
            _filename=filename,
            _group=group,
            _dataset=dataset,
            _dataset_kwds=dataset_kwds,
            _container_name=container_name,
            **kw,
        )

    @property
    def input(self):
        return self._input

    @property
    def filename(self):
        return self._filename

    @property
    def group(self):
        return self._group

    @property
    def dataset(self):
        return self._dataset

    @property
    def dataset_kwds(self):
        return self._dataset_kwds

    @property
    def axis_offsets(self):
        return self._axis_offsets

    @property
    def out_shape(self):
        return self._out_shape

    @property
    def container_name(self):
        return self._container_name

    @property
    def path(self):
        paths = []
        if self._group is not None:
            paths.append(self.group)
        paths.append(self.dataset)
        return "/".join(paths)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @classmethod
    def tile(cls, op):
        if has_unknown_shape(*op.inputs):
            yield
        in_tensor = op.input

        with open_file(op.filename, "w"):
            # create file if not exist
            pass

        nsplits = tuple([(0,) * len(ns) for ns in in_tensor.nsplits])
        if len(in_tensor.chunks) == 1:
            in_chunk = in_tensor.chunks[0]
            chunk_op = op.copy().reset_key()
            chunk_op._axis_offsets = (0,) * in_chunk.ndim
            chunk_op._out_shape = in_tensor.shape
            out_chunk = chunk_op.new_chunk(
                [in_chunk], shape=(0,) * in_chunk.ndim, index=in_chunk.index
            )
            new_op = op.copy()
            return new_op.new_tensors(
                op.inputs,
                shape=(0,) * in_tensor.ndim,
                nsplits=nsplits,
                chunks=[out_chunk],
            )

        container_name = f"{op.key}_{int(time.time() * 1000)}"

        out_chunks = []
        acc = [[0] + np.cumsum(ns).tolist() for ns in in_tensor.nsplits]
        chunk_op_keys = []
        for chunk in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._out_shape = in_tensor.shape
            chunk_op._container_name = container_name
            chunk_op._axis_offsets = tuple(
                acc[ax][i] for ax, i in enumerate(chunk.index)
            )
            out_chunk = chunk_op.new_chunk(
                [chunk], shape=(0,) * chunk.ndim, index=chunk.index
            )
            out_chunks.append(out_chunk)
            chunk_op_keys.append(out_chunk.op.key)

        ctx = get_context()
        ctx.create_remote_object(container_name, _HDF5Container, chunk_op_keys)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs, shape=(0,) * in_tensor.ndim, nsplits=nsplits, chunks=out_chunks
        )

    @classmethod
    def execute(cls, ctx, op: "TensorHDF5DataStore"):
        import h5py

        to_store = ctx[op.inputs[0].key]
        axis_offsets = op.axis_offsets

        container_name = op.container_name
        container: _HDF5Container = None
        if container_name:
            container = ctx.get_remote_object(container_name)
            container.acquire()
        try:
            with h5py.File(open_file(op.filename, mode="r+b"), mode="r+") as f:
                try:
                    ds = f[op.path]
                except KeyError:
                    ds = f.create_dataset(
                        op.path,
                        shape=op.out_shape,
                        dtype=to_store.dtype,
                        **op.dataset_kwds,
                    )
                ds[
                    tuple(
                        slice(offset, offset + size)
                        for offset, size in zip(axis_offsets, to_store.shape)
                    )
                ] = to_store
                ctx[op.outputs[0].key] = np.empty(
                    (0,) * to_store.ndim, dtype=to_store.dtype
                )
                if container:
                    container.mark_done(op.key)
        finally:
            if container:
                try:
                    container.release()
                    if container.is_done():
                        ctx.destroy_remote_object(container_name)
                except ActorNotExist:
                    # destroyed by other execution, just ignore
                    return


def tohdf5(hdf5_file, x, group=None, dataset=None, **kwds):
    import h5py

    x = astensor(x)
    if isinstance(hdf5_file, h5py.Dataset):
        filename = hdf5_file.file.filename
        group = hdf5_file.parent.name
        dataset = hdf5_file.name.rsplit("/", 1)[1]
    elif isinstance(hdf5_file, h5py.File):
        filename = hdf5_file.filename
        if dataset is None:
            raise ValueError("`dataset` should be provided")
    elif isinstance(hdf5_file, str):
        filename = hdf5_file
        if dataset is None:
            raise ValueError("`dataset` should be provided")
    else:
        raise TypeError(
            "`hdf5_file` passed has wrong type, "
            "expect str, h5py.File or h5py.Dataset, "
            f"got {type(hdf5_file)}"
        )

    op = TensorHDF5DataStore(
        filename=filename, group=group, dataset=dataset, dataset_kwds=kwds
    )
    return op(x)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...config import options
from ...core import OutputType
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameMin(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.MIN
    _func_name = "min"

    @property
    def is_atomic(self):
        return True


def min_series(df, axis=None, skipna=True, level=None, combine_size=None, method=None):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameMin(
        axis=axis,
        skipna=skipna,
        level=level,
        combine_size=combine_size,
        output_types=[OutputType.scalar],
        use_inf_as_na=use_inf_as_na,
        method=method,
    )
    return op(df)


def min_dataframe(
    df,
    axis=None,
    skipna=True,
    level=None,
    numeric_only=None,
    combine_size=None,
    method=None,
):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameMin(
        axis=axis,
        skipna=skipna,
        level=level,
        numeric_only=numeric_only,
        combine_size=combine_size,
        output_types=[OutputType.series],
        use_inf_as_na=use_inf_as_na,
        method=method,
    )
    return op(df)


def min_index(df, axis=None, skipna=True):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameMin(
        axis=axis,
        skipna=skipna,
        output_types=[OutputType.scalar],
        use_inf_as_na=use_inf_as_na,
    )
    return op(df)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
from typing import List, Tuple
from urllib.parse import urlparse

import numpy as np
import pandas as pd

try:
    import pyarrow as pa
    import pyarrow.parquet as pq
except ImportError:
    pa = None

try:
    import fastparquet
except ImportError:
    fastparquet = None

from ... import opcodes as OperandDef
from ...config import options
from ...lib.filesystem import file_size, get_fs, glob, open_file, FileSystem
from ...serialization.serializables import (
    AnyField,
    BoolField,
    DictField,
    ListField,
    StringField,
    Int32Field,
    Int64Field,
    BytesField,
)
from ...utils import is_object_dtype
from ..arrays import ArrowStringDtype
from ..operands import OutputType
from ..utils import (
    parse_index,
    to_arrow_dtypes,
    contain_arrow_dtype,
)
from .core import (
    IncrementalIndexDatasource,
    ColumnPruneSupportedDataSourceMixin,
    IncrementalIndexDataSourceMixin,
    merge_small_files,
)


PARQUET_MEMORY_SCALE = 15
STRING_FIELD_OVERHEAD = 50


def check_engine(engine):
    if engine == "auto":
        if pa is not None:
            return "pyarrow"
        elif fastparquet is not None:  # pragma: no cover
            return "fastparquet"
        else:  # pragma: no cover
            raise RuntimeError("Please install either pyarrow or fastparquet.")
    elif engine == "pyarrow":
        if pa is None:  # pragma: no cover
            raise RuntimeError("Please install pyarrow first.")
        return engine
    elif engine == "fastparquet":
        if fastparquet is None:  # pragma: no cover
            raise RuntimeError("Please install fastparquet first.")
        return engine
    else:  # pragma: no cover
        raise RuntimeError("Unsupported engine {} to read parquet.".format(engine))


def get_engine(engine):
    if engine == "pyarrow":
        return ArrowEngine()
    elif engine == "fastparquet":
        return FastpaquetEngine()
    else:  # pragma: no cover
        raise RuntimeError("Unsupported engine {}".format(engine))


class ParquetEngine:
    def get_row_num(self, f):
        raise NotImplementedError

    def read_dtypes(self, f, **kwargs):
        raise NotImplementedError

    def read_to_pandas(
        self, f, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        raise NotImplementedError

    def read_group_to_pandas(
        self, f, group_index, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        raise NotImplementedError

    def read_partitioned_to_pandas(
        self,
        f,
        partitions: "pq.ParquetPartitions",
        partition_keys: List[Tuple],
        columns=None,
        nrows=None,
        use_arrow_dtype=None,
        **kwargs,
    ):
        raw_df = self.read_to_pandas(
            f, columns=columns, nrows=nrows, use_arrow_dtype=use_arrow_dtype, **kwargs
        )
        for i, (name, index) in enumerate(partition_keys):
            dictionary = partitions[i].dictionary
            value = dictionary[index]
            raw_df[name] = pd.Categorical(
                np.full(raw_df.shape[0], value.as_py()), categories=dictionary.tolist()
            )
        return raw_df

    def read_partitioned_dtypes(self, fs: FileSystem, directory, storage_options):
        # As ParquetDataset will iterate all files,
        # here we just find one file to infer dtypes
        current_path = directory
        partition_cols = []
        while fs.isdir(current_path):
            _, dirs, files = next(fs.walk(current_path))
            dirs = [d for d in dirs if not d.startswith(".")]
            files = [f for f in files if not f.startswith(".")]
            if len(files) == 0:
                # directory as partition
                partition_cols.append(dirs[0].split("=", 1)[0])
                current_path = os.path.join(current_path, dirs[0])
            elif len(dirs) == 0:
                # parquet files in deepest directory
                current_path = os.path.join(current_path, files[0])
            else:  # pragma: no cover
                raise ValueError(
                    "Files and directories are mixed in an intermediate directory"
                )

        # current path is now a parquet file
        with open_file(current_path, storage_options=storage_options) as f:
            dtypes = self.read_dtypes(f)
        for partition in partition_cols:
            dtypes[partition] = pd.CategoricalDtype()
        return dtypes


def _parse_prefix(path):
    path_prefix = ""
    if isinstance(path, str):
        parsed_path = urlparse(path)
        if parsed_path.scheme:
            path_prefix = f"{parsed_path.scheme}://{parsed_path.netloc}"
    return path_prefix


class ArrowEngine(ParquetEngine):
    def get_row_num(self, f):
        file = pq.ParquetFile(f)
        return file.metadata.num_rows

    def read_dtypes(self, f, **kwargs):
        file = pq.ParquetFile(f)
        return file.schema_arrow.empty_table().to_pandas().dtypes

    @classmethod
    def _table_to_pandas(cls, t, nrows=None, use_arrow_dtype=None):
        if nrows is not None:
            t = t.slice(0, nrows)
        if use_arrow_dtype:
            df = t.to_pandas(types_mapper={pa.string(): ArrowStringDtype()}.get)
        else:
            df = t.to_pandas()
        return df

    def read_to_pandas(
        self, f, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        file = pq.ParquetFile(f)
        t = file.read(columns=columns, **kwargs)
        return self._table_to_pandas(t, nrows=nrows, use_arrow_dtype=use_arrow_dtype)

    def read_group_to_pandas(
        self, f, group_index, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        file = pq.ParquetFile(f)
        t = file.read_row_group(group_index, columns=columns, **kwargs)
        return self._table_to_pandas(t, nrows=nrows, use_arrow_dtype=use_arrow_dtype)


class FastpaquetEngine(ParquetEngine):
    def get_row_num(self, f):
        file = fastparquet.ParquetFile(f)
        return file.count()

    def read_dtypes(self, f, **kwargs):
        file = fastparquet.ParquetFile(f)
        dtypes_dict = file._dtypes()
        return pd.Series(dict((c, dtypes_dict[c]) for c in file.columns))

    def read_to_pandas(
        self, f, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        file = fastparquet.ParquetFile(f)
        df = file.to_pandas(columns, **kwargs)
        if nrows is not None:
            df = df.head(nrows)
        if use_arrow_dtype:
            df = df.astype(to_arrow_dtypes(df.dtypes).to_dict())
        return df


class DataFrameReadParquet(
    IncrementalIndexDatasource,
    ColumnPruneSupportedDataSourceMixin,
    IncrementalIndexDataSourceMixin,
):
    _op_type_ = OperandDef.READ_PARQUET

    path = AnyField("path")
    engine = StringField("engine")
    columns = ListField("columns")
    use_arrow_dtype = BoolField("use_arrow_dtype")
    groups_as_chunks = BoolField("groups_as_chunks")
    group_index = Int32Field("group_index")
    read_kwargs = DictField("read_kwargs")
    incremental_index = BoolField("incremental_index")
    storage_options = DictField("storage_options")
    is_partitioned = BoolField("is_partitioned")
    merge_small_files = BoolField("merge_small_files")
    merge_small_file_options = DictField("merge_small_file_options")
    # for chunk
    partitions = BytesField("partitions", default=None)
    partition_keys = ListField("partition_keys", default=None)
    num_group_rows = Int64Field("num_group_rows", default=None)
    # as read meta may be too time-consuming when number of files is large,
    # thus we only read first file to get row number and raw file size
    first_chunk_row_num = Int64Field("first_chunk_row_num")
    first_chunk_raw_bytes = Int64Field("first_chunk_raw_bytes")

    def get_columns(self):
        return self.columns

    def set_pruned_columns(self, columns, *, keep_order=None):
        self.columns = columns

    @classmethod
    def _to_arrow_dtypes(cls, dtypes, op):
        if (
            op.use_arrow_dtype is None
            and not op.gpu
            and options.dataframe.use_arrow_dtype
        ):  # pragma: no cover
            # check if use_arrow_dtype set on the server side
            dtypes = to_arrow_dtypes(dtypes)
        return dtypes

    @classmethod
    def _tile_partitioned(cls, op: "DataFrameReadParquet"):
        out_df = op.outputs[0]
        shape = (np.nan, out_df.shape[1])
        dtypes = cls._to_arrow_dtypes(out_df.dtypes, op)
        dataset = pq.ParquetDataset(op.path)

        path_prefix = _parse_prefix(op.path)

        chunk_index = 0
        out_chunks = []
        first_chunk_row_num, first_chunk_raw_bytes = None, None
        for i, piece in enumerate(dataset.pieces):
            chunk_op = op.copy().reset_key()
            chunk_op.path = chunk_path = path_prefix + piece.path
            chunk_op.partitions = pickle.dumps(dataset.partitions)
            chunk_op.partition_keys = piece.partition_keys
            if i == 0:
                first_chunk_raw_bytes = file_size(chunk_path, op.storage_options)
                first_chunk_row_num = piece.get_metadata().num_rows
            chunk_op.first_chunk_row_num = first_chunk_row_num
            chunk_op.first_chunk_raw_bytes = first_chunk_raw_bytes
            new_chunk = chunk_op.new_chunk(
                None,
                shape=shape,
                index=(chunk_index, 0),
                index_value=out_df.index_value,
                columns_value=out_df.columns_value,
                dtypes=dtypes,
            )
            out_chunks.append(new_chunk)
            chunk_index += 1

        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_df.shape[1],))
        return new_op.new_dataframes(
            None,
            out_df.shape,
            dtypes=dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile_no_partitioned(cls, op: "DataFrameReadParquet"):
        chunk_index = 0
        out_chunks = []
        out_df = op.outputs[0]

        dtypes = cls._to_arrow_dtypes(out_df.dtypes, op)
        shape = (np.nan, out_df.shape[1])

        path_prefix = ""
        if isinstance(op.path, (tuple, list)):
            paths = op.path
        elif get_fs(op.path, op.storage_options).isdir(op.path):
            parsed_path = urlparse(op.path)
            if parsed_path.scheme.lower() == "hdfs":
                path_prefix = f"{parsed_path.scheme}://{parsed_path.netloc}"
            paths = get_fs(op.path, op.storage_options).ls(op.path)
        else:
            paths = glob(op.path, storage_options=op.storage_options)

        first_chunk_row_num, first_chunk_raw_bytes = None, None
        for i, pth in enumerate(paths):
            pth = path_prefix + pth
            if i == 0:
                with open_file(pth, storage_options=op.storage_options) as f:
                    first_chunk_row_num = get_engine(op.engine).get_row_num(f)
                first_chunk_raw_bytes = file_size(
                    pth, storage_options=op.storage_options
                )

            if op.groups_as_chunks:
                num_row_groups = pq.ParquetFile(pth).num_row_groups
                for group_idx in range(num_row_groups):
                    chunk_op = op.copy().reset_key()
                    chunk_op.path = pth
                    chunk_op.group_index = group_idx
                    chunk_op.first_chunk_row_num = first_chunk_row_num
                    chunk_op.first_chunk_raw_bytes = first_chunk_raw_bytes
                    chunk_op.num_group_rows = num_row_groups
                    new_chunk = chunk_op.new_chunk(
                        None,
                        shape=shape,
                        index=(chunk_index, 0),
                        index_value=out_df.index_value,
                        columns_value=out_df.columns_value,
                        dtypes=dtypes,
                    )
                    out_chunks.append(new_chunk)
                    chunk_index += 1
            else:
                chunk_op = op.copy().reset_key()
                chunk_op.path = pth
                chunk_op.first_chunk_row_num = first_chunk_row_num
                chunk_op.first_chunk_raw_bytes = first_chunk_raw_bytes
                new_chunk = chunk_op.new_chunk(
                    None,
                    shape=shape,
                    index=(chunk_index, 0),
                    index_value=out_df.index_value,
                    columns_value=out_df.columns_value,
                    dtypes=dtypes,
                )
                out_chunks.append(new_chunk)
                chunk_index += 1

        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_df.shape[1],))
        return new_op.new_dataframes(
            None,
            out_df.shape,
            dtypes=dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile(cls, op: "DataFrameReadParquet"):
        if op.is_partitioned:
            tiled = cls._tile_partitioned(op)
        else:
            tiled = cls._tile_no_partitioned(op)
        if op.merge_small_files:
            tiled = [
                merge_small_files(tiled[0], **(op.merge_small_file_options or dict()))
            ]
        return tiled

    @classmethod
    def _execute_partitioned(cls, ctx, op: "DataFrameReadParquet"):
        out = op.outputs[0]
        engine = get_engine(op.engine)
        partitions = pickle.loads(op.partitions)
        with open_file(op.path, storage_options=op.storage_options) as f:
            ctx[out.key] = engine.read_partitioned_to_pandas(
                f,
                partitions,
                op.partition_keys,
                columns=op.columns,
                nrows=op.nrows,
                use_arrow_dtype=op.use_arrow_dtype,
                **op.read_kwargs or dict(),
            )

    @classmethod
    def execute(cls, ctx, op: "DataFrameReadParquet"):
        out = op.outputs[0]
        path = op.path

        if op.partitions is not None:
            return cls._execute_partitioned(ctx, op)

        engine = get_engine(op.engine)
        with open_file(path, storage_options=op.storage_options) as f:
            use_arrow_dtype = contain_arrow_dtype(out.dtypes)
            if op.groups_as_chunks:
                df = engine.read_group_to_pandas(
                    f,
                    op.group_index,
                    columns=op.columns,
                    nrows=op.nrows,
                    use_arrow_dtype=use_arrow_dtype,
                    **op.read_kwargs or dict(),
                )
            else:
                df = engine.read_to_pandas(
                    f,
                    columns=op.columns,
                    nrows=op.nrows,
                    use_arrow_dtype=use_arrow_dtype,
                    **op.read_kwargs or dict(),
                )

            ctx[out.key] = df

    @classmethod
    def estimate_size(cls, ctx, op: "DataFrameReadParquet"):
        first_chunk_row_num = op.first_chunk_row_num
        first_chunk_raw_bytes = op.first_chunk_raw_bytes
        raw_bytes = file_size(op.path, storage_options=op.storage_options)
        if op.num_group_rows:
            raw_bytes = (
                np.ceil(np.divide(raw_bytes, op.num_group_rows)).astype(np.int64).item()
            )

        estimated_row_num = (
            np.ceil(first_chunk_row_num * (raw_bytes / first_chunk_raw_bytes))
            .astype(np.int64)
            .item()
        )
        phy_size = raw_bytes * (op.memory_scale or PARQUET_MEMORY_SCALE)
        n_strings = len([dt for dt in op.outputs[0].dtypes if is_object_dtype(dt)])
        pd_size = phy_size + n_strings * estimated_row_num * STRING_FIELD_OVERHEAD
        ctx[op.outputs[0].key] = (pd_size, pd_size + phy_size)

    def __call__(self, index_value=None, columns_value=None, dtypes=None):
        self._output_types = [OutputType.dataframe]
        shape = (np.nan, len(dtypes))
        return self.new_dataframe(
            None,
            shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )


def read_parquet(
    path,
    engine: str = "auto",
    columns: list = None,
    groups_as_chunks: bool = False,
    use_arrow_dtype: bool = None,
    incremental_index: bool = False,
    storage_options: dict = None,
    memory_scale: int = None,
    merge_small_files: bool = True,
    merge_small_file_options: dict = None,
    **kwargs,
):
    """
    Load a parquet object from the file path, returning a DataFrame.

    Parameters
    ----------
    path : str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL.
        For file URLs, a host is expected. A local file could be:
        ``file://localhost/path/to/table.parquet``.
        A file URL can also be a path to a directory that contains multiple
        partitioned parquet files. Both pyarrow and fastparquet support
        paths to directories as well as file URLs. A directory path could be:
        ``file://localhost/path/to/tables``.
        By file-like object, we refer to objects with a ``read()`` method,
        such as a file handler (e.g. via builtin ``open`` function)
        or ``StringIO``.
    engine : {'auto', 'pyarrow', 'fastparquet'}, default 'auto'
        Parquet library to use. The default behavior is to try 'pyarrow',
        falling back to 'fastparquet' if 'pyarrow' is unavailable.
    columns : list, default=None
        If not None, only these columns will be read from the file.
    groups_as_chunks : bool, default False
        if True, each row group correspond to a chunk.
        if False, each file correspond to a chunk.
        Only available for 'pyarrow' engine.
    incremental_index: bool, default False
        If index_col not specified, ensure range index incremental,
        gain a slightly better performance if setting False.
    use_arrow_dtype: bool, default None
        If True, use arrow dtype to store columns.
    storage_options: dict, optional
        Options for storage connection.
    memory_scale: int, optional
        Scale that real memory occupation divided with raw file size.
    merge_small_files: bool, default True
        Merge small files whose size is small.
    merge_small_file_options: dict
        Options for merging small files
    **kwargs
        Any additional kwargs are passed to the engine.

    Returns
    -------
    Mars DataFrame
    """

    engine_type = check_engine(engine)
    engine = get_engine(engine_type)

    single_path = path[0] if isinstance(path, list) else path
    fs = get_fs(single_path, storage_options)
    is_partitioned = False
    if fs.isdir(single_path):
        paths = fs.ls(path)
        if all(fs.isdir(p) for p in paths):
            # If all are directories, it is read as a partitioned dataset.
            dtypes = engine.read_partitioned_dtypes(fs, path, storage_options)
            is_partitioned = True
        else:
            with fs.open(paths[0], mode="rb") as f:
                dtypes = engine.read_dtypes(f)
    else:
        if not isinstance(path, list):
            file_path = glob(path, storage_options=storage_options)[0]
        else:
            file_path = path[0]

        with open_file(file_path, storage_options=storage_options) as f:
            dtypes = engine.read_dtypes(f)

    if columns:
        dtypes = dtypes[columns]

    if use_arrow_dtype is None:
        use_arrow_dtype = options.dataframe.use_arrow_dtype
    if use_arrow_dtype:
        dtypes = to_arrow_dtypes(dtypes)

    index_value = parse_index(pd.RangeIndex(-1))
    columns_value = parse_index(dtypes.index, store_data=True)
    op = DataFrameReadParquet(
        path=path,
        engine=engine_type,
        columns=columns,
        groups_as_chunks=groups_as_chunks,
        use_arrow_dtype=use_arrow_dtype,
        read_kwargs=kwargs,
        incremental_index=incremental_index,
        storage_options=storage_options,
        is_partitioned=is_partitioned,
        memory_scale=memory_scale,
        merge_small_files=merge_small_files,
        merge_small_file_options=merge_small_file_options,
    )
    return op(index_value=index_value, columns_value=columns_value, dtypes=dtypes)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from pyarrow.fs import HadoopFileSystem as _ArrowHadoopFileSystem
    from .arrow import HadoopFileSystem

    del _ArrowHadoopFileSystem
except ImportError:  # pragma: no cover
    try:
        # pyarrow < 2.0.0
        from pyarrow import HadoopFileSystem
    except ImportError:
        HadoopFileSystem = None

from .core import register_filesystem


if HadoopFileSystem is not None:  # pragma: no branch
    register_filesystem("hdfs", HadoopFileSystem)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand
from ..array_utils import get_array_module, is_sparse_module


@arithmetic_operand(sparse_mode="unary")
class TensorI0(TensorUnaryOp):
    _op_type_ = OperandDef.I0
    _func_name = "i0"

    @classmethod
    def execute(cls, ctx, op):
        x = ctx[op.inputs[0].key]
        xp = get_array_module(x)
        res = xp.i0(x)
        if not is_sparse_module(xp):
            res = res.reshape(op.outputs[0].shape)
        ctx[op.outputs[0].key] = res


@infer_dtype(np.i0)
def i0(x, **kwargs):
    """
    Modified Bessel function of the first kind, order 0.

    Usually denoted :math:`I_0`.  This function does broadcast, but will *not*
    "up-cast" int dtype arguments unless accompanied by at least one float or
    complex dtype argument (see Raises below).

    Parameters
    ----------
    x : array_like, dtype float or complex
        Argument of the Bessel function.

    Returns
    -------
    out : Tensor, shape = x.shape, dtype = x.dtype
        The modified Bessel function evaluated at each of the elements of `x`.

    Raises
    ------
    TypeError: array cannot be safely cast to required type
        If argument consists exclusively of int dtypes.

    See Also
    --------
    scipy.special.iv, scipy.special.ive

    Notes
    -----
    We use the algorithm published by Clenshaw [1]_ and referenced by
    Abramowitz and Stegun [2]_, for which the function domain is
    partitioned into the two intervals [0,8] and (8,inf), and Chebyshev
    polynomial expansions are employed in each interval. Relative error on
    the domain [0,30] using IEEE arithmetic is documented [3]_ as having a
    peak of 5.8e-16 with an rms of 1.4e-16 (n = 30000).

    References
    ----------
    .. [1] C. W. Clenshaw, "Chebyshev series for mathematical functions", in
           *National Physical Laboratory Mathematical Tables*, vol. 5, London:
           Her Majesty's Stationery Office, 1962.
    .. [2] M. Abramowitz and I. A. Stegun, *Handbook of Mathematical
           Functions*, 10th printing, New York: Dover, 1964, pp. 379.
           http://www.math.sfu.ca/~cbm/aands/page_379.htm
    .. [3] http://kobesearch.cpan.org/htdocs/Math-Cephes/Math/Cephes.html

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.i0([0.]).execute()
    array([1.])
    >>> mt.i0([0., 1. + 2j]).execute()
    array([ 1.00000000+0.j        ,  0.18785373+0.64616944j])

    """
    op = TensorI0(**kwargs)
    return op(x)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import OutputType
from ..core.operand import (
    Operand,
    TileableOperandMixin,
    HasInput,
    ShuffleProxy,
    MapReduceOperand,
    Fuse,
)
from ..serialization.serializables import DataTypeField
from ..utils import calc_nsplits


class TensorOperandMixin(TileableOperandMixin):
    __slots__ = ()
    _op_module_ = "tensor"
    _output_type_ = OutputType.tensor

    def new_tensors(
        self,
        inputs,
        shape=None,
        dtype=None,
        order=None,
        chunks=None,
        nsplits=None,
        output_limit=None,
        kws=None,
        **kw
    ):
        return self.new_tileables(
            inputs,
            shape=shape,
            chunks=chunks,
            nsplits=nsplits,
            output_limit=output_limit,
            kws=kws,
            dtype=dtype,
            order=order,
            **kw
        )

    def new_tensor(self, inputs, shape, dtype=None, order=None, **kw):
        if getattr(self, "output_limit") != 1:
            raise TypeError("cannot new tensor with more than 1 outputs")
        return self.new_tensors(inputs, shape=shape, dtype=dtype, order=order, **kw)[0]

    @classmethod
    def concat_tileable_chunks(cls, tileable):
        from .merge.concatenate import TensorConcatenate

        tensor = tileable
        assert not tensor.is_coarse()

        op = TensorConcatenate(dtype=tensor.dtype)
        chunk = TensorConcatenate(dtype=tensor.dtype).new_chunk(
            tensor.chunks, shape=tensor.shape, index=(0,) * tileable.ndim
        )
        return op.new_tensor(
            [tensor],
            tensor.shape,
            chunks=[chunk],
            nsplits=tuple((s,) for s in tensor.shape),
        )

    @classmethod
    def create_tileable_from_chunks(cls, chunks, inputs=None, **kw):
        chunk_idx_to_shape = {c.index: c.shape for c in chunks}
        nsplits = calc_nsplits(chunk_idx_to_shape)
        shape = tuple(sum(ns) for ns in nsplits)
        op = chunks[0].op.copy().reset_key()
        return op.new_tensor(
            inputs,
            shape=shape,
            chunks=chunks,
            nsplits=nsplits,
            dtype=chunks[0].dtype,
            **kw
        )

    def get_fuse_op_cls(self, _):
        from .fuse import TensorFuseChunk

        return TensorFuseChunk


class TensorOperand(Operand):
    _output_type_ = OutputType.tensor

    dtype = DataTypeField("dtype", default=None)


class TensorHasInput(HasInput):
    _output_type_ = OutputType.tensor

    dtype = DataTypeField("dtype", default=None)


class TensorShuffleProxy(ShuffleProxy, TensorOperandMixin):
    _output_type_ = OutputType.tensor

    dtype = DataTypeField("dtype", default=None)

    @classmethod
    def execute(cls, ctx, op):
        pass


class TensorMapReduceOperand(MapReduceOperand):
    _output_type_ = OutputType.tensor

    dtype = DataTypeField("dtype", default=None)


class TensorFuse(Fuse):
    _output_type_ = OutputType.tensor

    dtype = DataTypeField("dtype", default=None)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Any

import numpy as np
import pandas as pd

from ...config import options
from ...core import OutputType
from ...core.context import Context
from ...lib.bloom_filter import BloomFilter
from ...serialization.serializables import (
    AnyField,
    Int64Field,
    Float64Field,
    StringField,
)
from ... import opcodes as OperandDef
from ...typing import TileableType
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import parse_index


class DataFrameBloomFilter(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_BLOOM_FILTER

    left_on = AnyField("left_on")
    right_on = AnyField("right_on")
    on = AnyField("on")
    # for build
    max_elements = Int64Field("max_elements")
    error_rate = Float64Field("error_rate")
    combine_size = Int64Field("combine_size")
    # chunk
    execution_stage = StringField("execution_stage", default=None)

    def __init__(self, execution_stage=None, **kwargs):
        if execution_stage in ["build", "union"]:
            output_types = [OutputType.object]
        else:
            output_types = [OutputType.dataframe]
        kwargs["_output_types"] = output_types
        super().__init__(execution_stage=execution_stage, **kwargs)

    def __call__(self, df1: TileableType, df2: TileableType):
        return self.new_tileable([df1, df2], **df1.params)

    @classmethod
    def tile(cls, op: "DataFrameBloomFilter"):
        df1, df2 = op.inputs
        # use df2's chunks to build bloom filter
        chunks = []
        for c in df2.chunks:
            build_op = DataFrameBloomFilter(
                on=op.right_on,
                max_elements=op.max_elements,
                error_rate=op.error_rate,
                execution_stage="build",
            )
            chunks.append(build_op.new_chunk(inputs=[c]))

        # union all chunk filters
        combine_size = op.combine_size
        while len(chunks) > combine_size:
            new_chunks = []
            for i in range(0, len(chunks), combine_size):
                chks = chunks[i : i + combine_size]
                if len(chks) == 1:
                    chk = chks[0]
                else:
                    union_op = DataFrameBloomFilter(execution_stage="union")
                    for j, c in enumerate(chks):
                        c._index = (j, 0)
                    chk = union_op.new_chunk(chks)
                new_chunks.append(chk)
            chunks = new_chunks
        if len(chunks) > 1:
            union_op = DataFrameBloomFilter(execution_stage="union")
            filter_chunk = union_op.new_chunk(chunks)
        else:
            filter_chunk = chunks[0]

        filter_chunk.is_broadcaster = True
        # filter df1
        out_chunks = []
        for chunk in df1.chunks:
            filter_op = DataFrameBloomFilter(on=op.left_on, execution_stage="filter")
            params = chunk.params.copy()
            params["shape"] = (np.nan, chunk.shape[1])
            params["index_value"] = parse_index(pd.RangeIndex(-1))
            out_chunks.append(filter_op.new_chunk([chunk, filter_chunk], **params))

        new_op = op.copy()
        df1_params = df1.params.copy()
        df1_params["chunks"] = out_chunks
        df1_params["nsplits"] = ((np.nan,) * len(out_chunks), df1.nsplits[1])
        return new_op.new_dataframes(op.inputs, **df1_params)

    @classmethod
    def _get_value(cls, value: Any):
        # value could be an element or a series, as BloomFilter
        # doesn't accept series, convert to list here
        if isinstance(value, pd.Series):
            return value.tolist()
        else:
            return value

    @classmethod
    def _filter_on_index(cls, on: Union[str, List, None], data: pd.DataFrame):
        if on is None:
            return True
        elif isinstance(on, str):
            return on not in data.columns
        else:
            assert isinstance(on, list)
            return any(c not in data.columns for c in on)

    @classmethod
    def _build_index_filter(cls, in_data: pd.DataFrame, op: "DataFrameBloomFilter"):
        if isinstance(in_data.index, pd.MultiIndex):
            index = in_data.index.get_level_values(op.on)
        else:
            index = in_data.index
        bloom_filter = BloomFilter(
            max_elements=op.max_elements, error_rate=op.error_rate
        )
        index.map(lambda v: bloom_filter.add(cls._get_value(v)))
        return bloom_filter

    @classmethod
    def _build_series_filter(cls, in_data: pd.Series, op: "DataFrameBloomFilter"):
        try:
            bloom_filter = BloomFilter(
                max_elements=op.max_elements, error_rate=op.error_rate
            )
            in_data[op.on].map(lambda v: bloom_filter.add(cls._get_value(v)))
        except TypeError:
            # has unhashable data, convert to str
            in_data = in_data.astype(str)
            bloom_filter = BloomFilter(
                max_elements=op.max_elements, error_rate=op.error_rate
            )
            in_data[op.on].map(lambda v: bloom_filter.add(cls._get_value(v)))
        return bloom_filter

    @classmethod
    def _build_dataframe_filter(cls, in_data: pd.DataFrame, op: "DataFrameBloomFilter"):
        try:
            bloom_filter = BloomFilter(
                max_elements=op.max_elements, error_rate=op.error_rate
            )
            in_data[op.on].apply(lambda v: bloom_filter.add(cls._get_value(v)), axis=1)
        except TypeError:
            # has unhashable data, convert to str
            in_data = in_data.astype(cls._convert_to_hashable_dtypes(in_data.dtypes))
            bloom_filter = BloomFilter(
                max_elements=op.max_elements, error_rate=op.error_rate
            )
            in_data[op.on].apply(lambda v: bloom_filter.add(cls._get_value(v)), axis=1)
        return bloom_filter

    @classmethod
    def _convert_to_hashable_dtypes(cls, dtypes: pd.Series):
        dtypes = dict(
            (name, dtype) if np.issubdtype(dtype, int) else (name, str)
            for name, dtype in dtypes.iteritems()
        )
        return dtypes

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "DataFrameBloomFilter"):
        if op.execution_stage == "build":
            on = op.on
            in_data = ctx[op.inputs[0].key]
            if cls._filter_on_index(on, in_data):
                bloom_filter = cls._build_index_filter(in_data, op)
            elif isinstance(on, str):
                bloom_filter = cls._build_series_filter(in_data, op)
            else:
                bloom_filter = cls._build_dataframe_filter(in_data, op)
            ctx[op.outputs[0].key] = bloom_filter
        elif op.execution_stage == "union":
            # union bloom filters
            filters = [ctx[inp.key] for inp in op.inputs]
            out = filters[0]
            for f in filters[1:]:
                out.union(f)
            ctx[op.outputs[0].key] = out
        elif op.execution_stage == "filter":
            on = op.on
            in_data = ctx[op.inputs[0].key]
            bloom_filter = ctx[op.inputs[1].key]
            if cls._filter_on_index(on, in_data):
                if isinstance(in_data.index, pd.MultiIndex):
                    idx = in_data.index.names.index(on)
                    ctx[op.outputs[0].key] = in_data[
                        in_data.index.map(lambda x: x[idx] in bloom_filter)
                    ]
                else:
                    ctx[op.outputs[0].key] = in_data[
                        in_data.index.map(lambda x: x in bloom_filter)
                    ]
            else:
                row_func = lambda row: cls._get_value(row) in bloom_filter
                if isinstance(on, str):
                    # series
                    try:
                        filtered = in_data[in_data[on].map(row_func)]
                    except TypeError:
                        converted_data = in_data.astype(str)
                        filtered = in_data[converted_data[on].map(row_func)]
                    ctx[op.outputs[0].key] = filtered
                else:
                    # dataframe
                    try:
                        filtered = in_data[in_data[on].apply(row_func, axis=1)]
                    except TypeError:
                        converted_data = in_data.astype(
                            cls._convert_to_hashable_dtypes(in_data.dtypes)
                        )
                        filtered = in_data[converted_data[on].apply(row_func, axis=1)]
                    ctx[op.outputs[0].key] = filtered

        else:  # pragma: no cover
            raise ValueError(f"Unknown execution stage: {op.execution_stage}")


def filter_by_bloom_filter(
    df1: TileableType,
    df2: TileableType,
    left_on: Union[str, List],
    right_on: Union[str, List],
    max_elements: int = 10000,
    error_rate: float = 0.1,
    combine_size: int = None,
):
    """
    Use bloom filter to filter DataFrame.

    Parameters
    ----------
    df1: DataFrame.
        DataFrame to be filtered.
    df2: DataFrame.
        Dataframe to build filter.
    left_on: str or list.
        Column(s) selected on df1.
    right_on: str or list.
        Column(s) selected on df2.
    max_elements: int
        How many elements you expect the filter to hold.
    error_rate: float
        error_rate defines accuracy.
    combine_size: int
        Combine size.

    Returns
    -------
    DataFrame
        Filtered df1.
    """
    if combine_size is None:
        combine_size = options.combine_size
    op = DataFrameBloomFilter(
        left_on=left_on,
        right_on=right_on,
        max_elements=max_elements,
        error_rate=error_rate,
        combine_size=combine_size,
    )
    return op(df1, df2)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import AnyField
from .core import TensorNoInput
from ..array_utils import create_array


class Scalar(TensorNoInput):
    """
    Operand represents scalar type.
    """

    _op_type_ = OperandDef.SCALAR

    _data = AnyField("data")

    def __init__(self, data=None, **kw):
        super().__init__(_data=data, **kw)

    @classmethod
    def tile(cls, op):
        chunk_op = op.copy().reset_key()
        chunk = chunk_op.new_chunk(None, shape=(), index=())
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs, op.outputs[0].shape, chunks=[chunk], nsplits=()
        )

    @property
    def data(self):
        return self._data

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        if chunk.ndim != 0:
            raise ValueError("Missing op for chunk")
        ctx[chunk.key] = create_array(op)("asarray", op.data)


def scalar(data, dtype=None, gpu=None):
    try:
        arr = np.array(data, dtype=dtype)
        op = Scalar(arr, dtype=arr.dtype, gpu=gpu)
        shape = ()
        return op(shape)
    except ValueError:
        raise TypeError(f"Expect scalar, got: {data}")



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict, List, Type

from ....typing import BandType
from ..core import _CommonMeta


class AbstractMetaStore(ABC):
    name = None

    def __init__(self, session_id: str, **kw):
        # make sure all meta types registered
        from .. import metas

        del metas

        self._session_id = session_id

    @classmethod
    @abstractmethod
    async def create(cls, config) -> Dict:
        """
        Create a meta store. Do some initialization work.
        For instance, for database backend,
        db files including tables may be created first.
        This should be done when service starting.

        Parameters
        ----------
        config : dict
            config.

        Returns
        -------
        kwargs : dict
            kwargs to create a meta store.
        """

    @abstractmethod
    async def set_meta(self, object_id: str, meta: _CommonMeta):
        """
        Set meta.

        Parameters
        ----------
        object_id : str
            Object ID.
        meta : _CommonMeta
            Meta.
        """

    @abstractmethod
    async def get_meta(
        self, object_id: str, fields: List[str] = None, error="raise"
    ) -> Dict:
        """
        Get meta.

        Parameters
        ----------
        object_id : str
            Object ID.
        fields : list
            Fields to filter, if not provided, get all fields.
        error : str
            'raise' or 'ignore'

        Returns
        -------
        meta: dict
            Meta.
        """

    @abstractmethod
    async def del_meta(self, object_id: str):
        """
        Delete meta.

        Parameters
        ----------
        object_id : str
            Object ID.
        """

    @abstractmethod
    async def add_chunk_bands(self, object_id: str, bands: List[BandType]):
        """
        Add band to chunk.

        Parameters
        ----------
        object_id : str
            Object ID.
        bands : List[BandType]
            Band of chunk to add, shall be tuple of (worker, band).
        """

    @abstractmethod
    async def remove_chunk_bands(self, object_id: str, bands: List[BandType]):
        """
        Remove bands from chunk.

        Parameters
        ----------
        object_id : str
            Object ID.
        bands : List[BandType]
            Bands of chunk to remove, shall be tuple of (worker, band).
        """

    @abstractmethod
    async def get_band_chunks(self, band: BandType) -> List[str]:
        """Get chunks key of band"""


_meta_store_types: Dict[str, Type[AbstractMetaStore]] = dict()


def register_meta_store(meta_store: Type[AbstractMetaStore]):
    _meta_store_types[meta_store.name] = meta_store
    return meta_store


def get_meta_store(meta_store_name: str) -> Type[AbstractMetaStore]:
    return _meta_store_types[meta_store_name]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import List, Dict, Tuple, Union, Iterable

from ...core import Tileable, Chunk
from ...serialization.core import buffered
from ...serialization.serializables import Serializable, DictField, ListField, BoolField
from ...serialization.serializables.core import SerializableSerializer
from ...utils import tokenize
from .core import DAG


class EntityGraph(DAG, metaclass=ABCMeta):
    @property
    @abstractmethod
    def results(self):
        """
        Return result tileables or chunks.

        Returns
        -------
        results
        """

    @results.setter
    @abstractmethod
    def results(self, new_results):
        """
        Set result tileables or chunks.

        Parameters
        ----------
        new_results

        Returns
        -------

        """

    def copy(self) -> "EntityGraph":
        graph = super().copy()
        graph.results = self.results.copy()
        return graph


class TileableGraph(EntityGraph, Iterable[Tileable]):
    _result_tileables: List[Tileable]
    # logic key is a unique and deterministic key for `TileableGraph`. For
    # multiple runs the logic key will remain same if the computational logic
    # doesn't change. And it can be used to some optimization when running a
    # same `execute`, like HBO.
    _logic_key: str

    def __init__(self, result_tileables: List[Tileable] = None):
        super().__init__()
        self._result_tileables = result_tileables

    @property
    def result_tileables(self):
        return self._result_tileables

    @property
    def results(self):
        return self._result_tileables

    @results.setter
    def results(self, new_results):
        self._result_tileables = new_results

    @property
    def logic_key(self):
        if not hasattr(self, "_logic_key") or self._logic_key is None:
            token_keys = []
            for node in self.bfs():
                token_keys.append(
                    tokenize(node.op.get_logic_key(), **node.extra_params)
                    if node.extra_params
                    else node.op.get_logic_key()
                )
            self._logic_key = tokenize(*token_keys)
        return self._logic_key


class ChunkGraph(EntityGraph, Iterable[Chunk]):
    _result_chunks: List[Chunk]

    def __init__(self, result_chunks: List[Chunk] = None):
        super().__init__()
        self._result_chunks = result_chunks

    @property
    def result_chunks(self):
        return self._result_chunks

    @property
    def results(self):
        return self._result_chunks

    @results.setter
    def results(self, new_results):
        self._result_chunks = new_results


class SerializableGraph(Serializable):
    _is_chunk = BoolField("is_chunk")
    # TODO(qinxuye): remove this logic when we handle fetch elegantly,
    # now, the node in the graph and inputs for operand may be inconsistent,
    # for example, an operand's inputs may be chunks,
    # but in the graph, the predecessors are all fetch chunks,
    # we serialize the fetch chunks first to make sure when operand's inputs
    # are serialized, they will just be marked as serialized and skip serialization.
    _fetch_nodes = ListField("fetch_nodes")
    _nodes = DictField("nodes")
    _predecessors = DictField("predecessors")
    _successors = DictField("successors")
    _results = ListField("results")

    @classmethod
    def from_graph(cls, graph: Union[TileableGraph, ChunkGraph]) -> "SerializableGraph":
        from ..operand import Fetch

        is_chunk = isinstance(graph, ChunkGraph)
        return SerializableGraph(
            _is_chunk=is_chunk,
            _fetch_nodes=[chunk for chunk in graph if isinstance(chunk.op, Fetch)],
            _nodes=graph._nodes,
            _predecessors=graph._predecessors,
            _successors=graph._successors,
            _results=graph.results,
        )

    def to_graph(self) -> Union[TileableGraph, ChunkGraph]:
        graph_cls = ChunkGraph if self._is_chunk else TileableGraph
        graph = graph_cls(self._results)
        graph._nodes.update(self._nodes)
        graph._predecessors.update(self._predecessors)
        graph._successors.update(self._successors)
        return graph


class GraphSerializer(SerializableSerializer):
    @buffered
    def serial(self, obj: Union[TileableGraph, ChunkGraph], context: Dict):
        serializable_graph = SerializableGraph.from_graph(obj)
        return (), [serializable_graph], False

    def deserial(
        self, serialized: Tuple, context: Dict, subs: List
    ) -> Union[TileableGraph, ChunkGraph]:
        serializable_graph: SerializableGraph = subs[0]
        return serializable_graph.to_graph()


GraphSerializer.register(EntityGraph)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, List, Optional, Tuple

try:
    import xgboost
except ImportError:
    xgboost = None

from .dmatrix import MarsDMatrix


XGBScikitLearnBase = None
if xgboost:

    class XGBScikitLearnBase(xgboost.XGBModel):
        """
        Base class for implementing scikit-learn interface
        """

        def fit(
            self,
            X,
            y,
            sample_weights=None,
            eval_set=None,
            sample_weight_eval_set=None,
            **kw,
        ):
            """
            Fit the regressor.
            Parameters
            ----------
            X : array_like
                Feature matrix
            y : array_like
                Labels
            sample_weight : array_like
                instance weights
            eval_set : list, optional
                A list of (X, y) tuple pairs to use as validation sets, for which
                metrics will be computed.
                Validation metrics will help us track the performance of the model.
            sample_weight_eval_set : list, optional
                A list of the form [L_1, L_2, ..., L_n], where each L_i is a list
                of group weights on the i-th validation set.
            """
            raise NotImplementedError

        def predict(self, data, **kw):
            """
            Predict with `data`.

            Parameters
            ----------
              data: data that can be used to perform prediction
            Returns
            -------
            prediction : mars.tensor.Tensor
            """
            raise NotImplementedError

    def wrap_evaluation_matrices(
        missing: float,
        X: Any,
        y: Any,
        sample_weight: Optional[Any],
        base_margin: Optional[Any],
        eval_set: Optional[List[Tuple[Any, Any]]],
        sample_weight_eval_set: Optional[List[Any]],
        base_margin_eval_set: Optional[List[Any]],
        label_transform: Callable = lambda x: x,
    ) -> Tuple[Any, Optional[List[Tuple[Any, str]]]]:
        """Convert array_like evaluation matrices into DMatrix.  Perform validation on the way."""
        train_dmatrix = MarsDMatrix(
            data=X,
            label=label_transform(y),
            weight=sample_weight,
            base_margin=base_margin,
            missing=missing,
        )

        n_validation = 0 if eval_set is None else len(eval_set)

        def validate_or_none(meta: Optional[List], name: str) -> List:
            if meta is None:
                return [None] * n_validation
            if len(meta) != n_validation:
                raise ValueError(
                    f"{name}'s length does not equal `eval_set`'s length, "
                    + f"expecting {n_validation}, got {len(meta)}"
                )
            return meta

        if eval_set is not None:
            sample_weight_eval_set = validate_or_none(
                sample_weight_eval_set, "sample_weight_eval_set"
            )
            base_margin_eval_set = validate_or_none(
                base_margin_eval_set, "base_margin_eval_set"
            )

            evals = []
            for i, (valid_X, valid_y) in enumerate(eval_set):
                # Skip the duplicated entry.
                if all(
                    (
                        valid_X is X,
                        valid_y is y,
                        sample_weight_eval_set[i] is sample_weight,
                        base_margin_eval_set[i] is base_margin,
                    )
                ):
                    evals.append(train_dmatrix)
                else:
                    m = MarsDMatrix(
                        data=valid_X,
                        label=label_transform(valid_y),
                        weight=sample_weight_eval_set[i],
                        base_margin=base_margin_eval_set[i],
                        missing=missing,
                    )
                    evals.append(m)
            nevals = len(evals)
            eval_names = [f"validation_{i}" for i in range(nevals)]
            evals = list(zip(evals, eval_names))
        else:
            if any(
                meta is not None
                for meta in [
                    sample_weight_eval_set,
                    base_margin_eval_set,
                ]
            ):
                raise ValueError(
                    "`eval_set` is not set but one of the other evaluation meta info is "
                    "not None."
                )
            evals = []

        return train_dmatrix, evals



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType, get_output_types, recursive_tile
from ...serialization.serializables import DictField, Int64Field, BoolField
from ...utils import pd_release_version
from ..core import IndexValue
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import build_concatenated_rows_frame, parse_index

_pandas_enable_negative = pd_release_version >= (1, 4, 0)


class GroupByHead(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.GROUPBY_HEAD
    _op_module_ = "dataframe.groupby"

    row_count = Int64Field("row_count")
    groupby_params = DictField("groupby_params")
    enable_negative = BoolField("enable_negative")

    def __call__(self, groupby):
        df = groupby
        while df.op.output_types[0] not in (OutputType.dataframe, OutputType.series):
            df = df.inputs[0]

        selection = groupby.op.groupby_params.pop("selection", None)
        if df.ndim > 1 and selection:
            if isinstance(selection, tuple) and selection not in df.dtypes:
                selection = list(selection)

            result_df = df[selection]
        else:
            result_df = df

        self._output_types = (
            [OutputType.dataframe] if result_df.ndim == 2 else [OutputType.series]
        )

        params = result_df.params
        params["shape"] = (np.nan,) + result_df.shape[1:]
        if isinstance(df.index_value.value, IndexValue.RangeIndex):
            params["index_value"] = parse_index(pd.RangeIndex(-1), df.key)

        return self.new_tileable([df], **params)

    @classmethod
    def tile(cls, op: "GroupByHead"):
        in_df = op.inputs[0]
        groupby_params = op.groupby_params.copy()
        selection = groupby_params.pop("selection", None)

        enable_negative = _pandas_enable_negative and op.enable_negative

        if len(in_df.shape) > 1:
            in_df = build_concatenated_rows_frame(in_df)
        out_df = op.outputs[0]

        # when row_count is not positive and pandas does not support negative head,
        #  or there is only one chunk, tile with a single chunk
        if (not enable_negative and op.row_count <= 0) or len(in_df.chunks) <= 1:
            row_num = 0 if not enable_negative and op.row_count <= 0 else np.nan
            new_shape = (row_num,)
            new_nsplits = ((row_num,),)
            if out_df.ndim > 1:
                new_shape += (out_df.shape[1],)
                new_nsplits += ((out_df.shape[1],),)

            c = in_df.chunks[0]
            chunk_op = op.copy().reset_key()
            params = out_df.params
            params["shape"] = new_shape
            params["index"] = (0,) * out_df.ndim
            out_chunk = chunk_op.new_chunk([c], **params)

            tileable_op = op.copy().reset_key()
            return tileable_op.new_tileables(
                [in_df], nsplits=new_nsplits, chunks=[out_chunk], **params
            )

        if in_df.ndim > 1 and selection:
            if isinstance(selection, tuple) and selection not in in_df.dtypes:
                selection = list(selection)

            if not isinstance(selection, list):
                pre_selection = [selection]
            else:
                pre_selection = list(selection)

            if isinstance(groupby_params.get("by"), list):
                pre_selection += [
                    el for el in groupby_params["by"] if el not in pre_selection
                ]

            if len(pre_selection) != in_df.shape[1]:
                in_df = yield from recursive_tile(in_df[pre_selection])

        # generate pre chunks
        if op.row_count < 0:
            # when we have negative row counts, pre-groupby optimization is not possible
            pre_chunks = in_df.chunks
        else:
            pre_chunks = []
            for c in in_df.chunks:
                pre_op = op.copy().reset_key()
                pre_op._output_types = get_output_types(c)
                pre_op.groupby_params = op.groupby_params.copy()
                pre_op.groupby_params.pop("selection", None)
                params = c.params
                params["shape"] = (np.nan,) + c.shape[1:]
                pre_chunks.append(pre_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        new_op._output_types = get_output_types(in_df)
        new_nsplits = ((np.nan,) * len(in_df.nsplits[0]),) + in_df.nsplits[1:]
        pre_tiled = new_op.new_tileable(
            [in_df], chunks=pre_chunks, nsplits=new_nsplits, **in_df.params
        )

        # generate groupby
        grouped = yield from recursive_tile(pre_tiled.groupby(**groupby_params))
        if selection:
            grouped = yield from recursive_tile(grouped[selection])

        # generate post chunks
        post_chunks = []
        for c in grouped.chunks:
            post_op = op.copy().reset_key()
            post_op.groupby_params = op.groupby_params.copy()
            post_op.groupby_params.pop("selection", None)
            if op.output_types[0] == OutputType.dataframe:
                index = c.index
            else:
                index = (c.index[0],)
            params = out_df.params
            params["index"] = index
            post_chunks.append(post_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        new_nsplits = ((np.nan,) * len(in_df.nsplits[0]),)
        if out_df.ndim > 1:
            new_nsplits += ((out_df.shape[1],),)
        return new_op.new_tileables(
            [in_df], chunks=post_chunks, nsplits=new_nsplits, **out_df.params
        )

    @classmethod
    def execute(cls, ctx, op: "GroupByHead"):
        in_data = ctx[op.inputs[0].key]

        params = op.groupby_params.copy()
        selection = params.pop("selection", None)

        if hasattr(in_data, "groupby"):
            grouped = in_data.groupby(**params)
        else:
            grouped = in_data

        if selection:
            grouped = grouped[selection]

        result = grouped.head(op.row_count)
        if not op.enable_negative and op.row_count < 0:
            result = result.iloc[:0]
        ctx[op.outputs[0].key] = result


def head(groupby, n=5):
    """
    Return first n rows of each group.

    Similar to ``.apply(lambda x: x.head(n))``, but it returns a subset of rows
    from the original Series or DataFrame with original index and order preserved
    (``as_index`` flag is ignored).

    Does not work for negative values of `n`.

    Returns
    -------
    Series or DataFrame

    See Also
    --------
    Series.groupby
    DataFrame.groupby

    Examples
    --------

    >>> import mars.dataframe as md
    >>> df = md.DataFrame([[1, 2], [1, 4], [5, 6]],
    ...                   columns=['A', 'B'])
    >>> df.groupby('A').head(1).execute()
       A  B
    0  1  2
    2  5  6
    >>> df.groupby('A').head(-1).execute()
    Empty DataFrame
    Columns: [A, B]
    Index: []
    """
    groupby_params = groupby.op.groupby_params.copy()
    groupby_params.pop("as_index", None)

    op = GroupByHead(
        row_count=n,
        groupby_params=groupby_params,
        enable_negative=_pandas_enable_negative,
    )
    return op(groupby)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import itertools

import numpy as np

from ... import opcodes
from ...core import ENTITY_TYPE, get_output_types, recursive_tile
from ...serialization.serializables import (
    BoolField,
    AnyField,
    Int8Field,
    Int64Field,
    Float64Field,
    KeyField,
)
from ...tensor import searchsorted
from ...tensor.base import TensorMapChunk
from ...tensor.merge import TensorConcatenate
from ...tensor.random import RandomState as TensorRandomState, RandomStateField
from ...tensor.utils import normalize_chunk_sizes, gen_random_seeds
from ...utils import has_unknown_shape, ceildiv
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import validate_axis, parse_index


class DataFrameSample(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.RAND_SAMPLE

    _size = Int64Field("size")
    _frac = Float64Field("frac")
    _replace = BoolField("replace")
    _weights = AnyField("weights")
    _axis = Int8Field("axis")
    _seed = Int64Field("seed")
    _random_state = RandomStateField("random_state")
    _always_multinomial = BoolField("always_multinomial")

    # for chunks
    # num of instances for chunks
    _chunk_samples = KeyField("chunk_samples")

    def __init__(
        self,
        size=None,
        frac=None,
        replace=None,
        weights=None,
        seed=None,
        axis=None,
        random_state=None,
        always_multinomial=None,
        chunk_samples=None,
        **kw
    ):
        super().__init__(
            _size=size,
            _frac=frac,
            _replace=replace,
            _weights=weights,
            _seed=seed,
            _axis=axis,
            _random_state=random_state,
            _always_multinomial=always_multinomial,
            _chunk_samples=chunk_samples,
            **kw
        )

    @property
    def size(self):
        return self._size

    @property
    def frac(self):
        return self._frac

    @property
    def replace(self):
        return self._replace

    @property
    def weights(self):
        return self._weights

    @property
    def seed(self):
        return self._seed

    @property
    def axis(self):
        return self._axis

    @property
    def random_state(self):
        if self._random_state is None:
            self._random_state = np.random.RandomState(self.seed)
        return self._random_state

    @property
    def always_multinomial(self):
        return self._always_multinomial

    @property
    def chunk_samples(self):
        return self._chunk_samples

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        it = iter(inputs)
        next(it)
        if isinstance(self.weights, ENTITY_TYPE):
            self._weights = next(it)
        if isinstance(self.chunk_samples, ENTITY_TYPE):
            self._chunk_samples = next(it)

    def __call__(self, df):
        params = df.params
        new_shape = list(df.shape)

        if self.frac is not None and not np.isnan(df.shape[self.axis]):
            self._size = int(self.frac * df.shape[self.axis])
            self._frac = None

        if self.size is not None:
            new_shape[self.axis] = self.size
        params["shape"] = tuple(new_shape)
        params["index_value"] = parse_index(df.index_value.to_pandas()[:0])

        input_dfs = [df]
        if isinstance(self.weights, ENTITY_TYPE):
            input_dfs.append(self.weights)

        self._output_types = get_output_types(df)
        return self.new_tileable(input_dfs, **params)

    @classmethod
    def _tile_one_chunk(cls, op: "DataFrameSample", in_df, weights):
        out = op.outputs[0]

        input_dfs = [in_df]
        if isinstance(weights, ENTITY_TYPE):
            input_dfs.append(weights)

        params = out.params
        chunk_op = op.copy().reset_key()
        if isinstance(weights, ENTITY_TYPE):
            chunk_op._weights = weights
        params["index"] = (0,) * out.ndim
        chunk = chunk_op.new_chunk([c.chunks[0] for c in input_dfs], **params)

        df_op = op.copy().reset_key()
        return df_op.new_tileables(
            input_dfs, chunks=[chunk], nsplits=((s,) for s in out.shape), **params
        )

    @classmethod
    def _tile_multinomial(cls, op: "DataFrameSample", in_df, weights):
        out_data = op.outputs[0]
        input_dfs = [in_df]
        size = op.size

        weight_chunks = itertools.repeat(None)
        if isinstance(op.weights, ENTITY_TYPE):
            input_dfs.append(weights)
            weight_chunks = weights.chunks

        chunks = []
        new_nsplits = list(in_df.nsplits)
        rs = op.random_state
        seeds = gen_random_seeds(len(in_df.chunks), op.random_state)
        if weights is None:
            # weights is None, use nsplits to sample num of instances for each chunk
            probs = np.array(in_df.nsplits[op.axis])
            probs = 1.0 * probs / probs.sum()
            chunk_sizes = rs.multinomial(size, probs)
            new_nsplits[op.axis] = tuple(int(s) for s in chunk_sizes if s > 0)

            chunk_idx = 0
            for data_chunk, chunk_size, seed in zip(in_df.chunks, chunk_sizes, seeds):
                if chunk_size == 0:
                    continue

                chunk_op = op.copy().reset_key()
                chunk_op._random_state = None
                chunk_op._seed = seed
                chunk_op._size = int(chunk_size)

                params = data_chunk.params
                params["index_value"] = parse_index(
                    params["index_value"].to_pandas()[:0]
                )
                new_shape = list(data_chunk.shape)
                new_shape[op.axis] = int(chunk_size)
                params["shape"] = tuple(new_shape)

                idx_list = [0] * data_chunk.ndim
                idx_list[op.axis] = chunk_idx
                params["index"] = tuple(idx_list)

                chunks.append(chunk_op.new_chunk([data_chunk], **params))
                chunk_idx += 1
        else:
            mn_seed = gen_random_seeds(1, op.random_state)[0]

            # weights is specified, use weights to sample num of instances for each chunk
            chunk_weights = yield from recursive_tile(
                weights.to_tensor().map_chunk(lambda x: x.sum(keepdims=True))
            )
            chunk_weights_chunk = TensorConcatenate(
                dtype=chunk_weights.dtype
            ).new_chunk(
                chunk_weights.chunks, shape=(len(chunk_weights.chunks),), index=(0,)
            )
            chunk_samples = TensorMapChunk(
                func=lambda x: np.random.RandomState(mn_seed).multinomial(
                    size, x / x.sum()
                )
            ).new_chunk(
                [chunk_weights_chunk], shape=(len(chunk_weights.chunks),), index=(0,)
            )
            new_nsplits[op.axis] = (np.nan,) * len(chunk_weights.chunks)
            for chunk_idx, (data_chunk, weight_chunk, seed) in enumerate(
                zip(in_df.chunks, weight_chunks, seeds)
            ):
                input_chunks = [data_chunk]

                chunk_op = op.copy().reset_key()
                chunk_op._size = None
                chunk_op._random_state = None
                chunk_op._seed = seed
                chunk_op._chunk_samples = chunk_samples
                if weight_chunk is not None:
                    chunk_op._weights = weight_chunk
                    input_chunks.append(weight_chunk)

                params = data_chunk.params
                params["index_value"] = parse_index(
                    params["index_value"].to_pandas()[:0]
                )
                new_shape = list(data_chunk.shape)
                new_shape[op.axis] = np.nan
                params["shape"] = tuple(new_shape)

                idx_list = [0] * data_chunk.ndim
                idx_list[op.axis] = chunk_idx
                params["index"] = tuple(idx_list)

                chunks.append(
                    chunk_op.new_chunk(input_chunks + [chunk_samples], **params)
                )

        params = out_data.params
        new_shape = list(in_df.shape)
        new_shape[op.axis] = size
        params["shape"] = tuple(new_shape)

        df_op = op.copy().reset_key()
        return df_op.new_tileables(
            input_dfs, chunks=chunks, nsplits=tuple(new_nsplits), **params
        )

    @classmethod
    def _tile_reservoirs(cls, op: "DataFrameSample", in_df, weights):
        out_data = op.outputs[0]
        input_dfs = [in_df]
        size = op.size

        weight_chunks = itertools.repeat(None)
        if isinstance(weights, ENTITY_TYPE):
            input_dfs.append(weights)
            weight_chunks = weights.chunks

        if any(cs < size for cs in in_df.nsplits[op.axis]):
            # make sure all chunk > m
            n_records = in_df.shape[op.axis]
            n_chunk = min(max(ceildiv(n_records, size), 1), in_df.chunk_shape[0])
            chunk_size = ceildiv(in_df.shape[op.axis], n_chunk)
            chunk_sizes = list(normalize_chunk_sizes(n_records, chunk_size)[0])
            if chunk_sizes[-1] < size and len(chunk_sizes) > 1:
                # the last chunk may still less than m
                # merge it into previous one
                chunk_sizes[-2] += chunk_sizes[-1]
                chunk_sizes = chunk_sizes[:-1]
            in_df = yield from recursive_tile(in_df.rechunk({0: tuple(chunk_sizes)}))
            if isinstance(weights, ENTITY_TYPE):
                weights = yield from recursive_tile(
                    weights.rechunk({0: tuple(chunk_sizes)})
                )
            if len(chunk_sizes) == 1:
                return cls._tile_one_chunk(op, in_df, weights)

        # for each chunk in a, do regular sampling
        sampled_chunks = []
        seeds = gen_random_seeds(len(in_df.chunks), op.random_state)
        for data_chunk, weights_chunk, seed in zip(in_df.chunks, weight_chunks, seeds):
            input_chunks = [data_chunk]

            chunk_op = op.copy().reset_key()
            chunk_op._random_state = None
            chunk_op._seed = seed
            if isinstance(op.weights, ENTITY_TYPE):
                input_chunks.append(weights_chunk)
                chunk_op._weights = weights_chunk

            params = data_chunk.params
            new_shape = list(data_chunk.shape)
            new_shape[op.axis] = size
            params["shape"] = tuple(new_shape)
            sampled_chunks.append(chunk_op.new_chunk(input_chunks, **params))

        # generate a random variable for samples in every chunk
        state = TensorRandomState.from_numpy(op.random_state)
        indices = state.rand(size)

        if weights is None:
            # weights not specified, use nsplits to calculate cumulative probability
            # to distribute samples in each chunk
            cum_offsets = np.cumsum(in_df.nsplits[op.axis])
            cum_offsets = cum_offsets * 1.0 / cum_offsets[-1]
        else:
            # weights specified, use weights to calculate cumulative probability
            # to distribute samples in each chunk
            chunk_weights = yield from recursive_tile(
                weights.to_tensor().map_chunk(lambda x: x.sum(keepdims=True))
            )
            chunk_weights_chunk = TensorConcatenate(
                dtype=chunk_weights.dtype
            ).new_chunk(
                chunk_weights.chunks, shape=(len(chunk_weights.chunks),), index=(0,)
            )

            cum_chunk = TensorMapChunk(func=lambda x: (x / x.sum()).cumsum()).new_chunk(
                [chunk_weights_chunk], shape=(len(chunk_weights.chunks),), index=(0,)
            )
            cum_offsets = TensorMapChunk(func=cum_chunk.op.func).new_tensor(
                [weights],
                chunks=[cum_chunk],
                nsplits=((s,) for s in cum_chunk.shape),
                **cum_chunk.params
            )

        index_chunks = []
        # seek which chunk the final sample will select
        chunk_sel = yield from recursive_tile(
            searchsorted(cum_offsets, indices, side="right")
        )
        # for every chunk, select samples with bool indexing
        for idx, sampled_chunk in enumerate(sampled_chunks):
            chunk_index = chunk_sel.map_chunk(
                func=lambda x, i: x == i, args=(idx,), elementwise=True, dtype=bool
            )
            sampled_df_op = sampled_chunk.op.copy().reset_key()
            sampled_chunk._index = (0,) * sampled_chunk.ndim
            sampled_df = sampled_df_op.new_tileable(
                input_dfs,
                chunks=[sampled_chunk],
                nsplits=((s,) for s in sampled_chunk.shape),
                **sampled_chunk.params
            )
            index_chunk = (
                yield from recursive_tile(sampled_df.iloc[chunk_index])
            ).chunks[0]

            chunk_idx = [0] * sampled_chunk.ndim
            chunk_idx[op.axis] = idx
            index_chunk._index = tuple(chunk_idx)
            index_chunks.append(index_chunk)

        params = out_data.params
        new_shape = list(in_df.shape)
        new_shape[op.axis] = size
        params["shape"] = tuple(new_shape)

        new_nsplits = list(in_df.nsplits)
        new_nsplits[op.axis] = (np.nan,) * len(index_chunks)

        df_op = op.copy().reset_key()
        return df_op.new_tileables(
            input_dfs, chunks=index_chunks, nsplits=tuple(new_nsplits), **params
        )

    @classmethod
    def tile(cls, op: "DataFrameSample"):
        if has_unknown_shape(*op.inputs):
            yield

        in_df = op.inputs[0]
        if in_df.ndim == 2:
            in_df = yield from recursive_tile(
                in_df.rechunk({(1 - op.axis): (in_df.shape[1 - op.axis],)})
            )

        if op.size is None:
            op._size = int(op.frac * in_df.shape[op.axis])

        weights = op.weights
        if isinstance(weights, ENTITY_TYPE):
            weights = yield from recursive_tile(
                weights.rechunk({0: in_df.nsplits[op.axis]})
            )
        elif in_df.ndim > 1 and weights in in_df.dtypes.index:
            weights = yield from recursive_tile(in_df[weights])

        if len(in_df.chunks) == 1:
            return cls._tile_one_chunk(op, in_df, weights)

        if op.replace or op.always_multinomial:
            return (yield from cls._tile_multinomial(op, in_df, weights))
        else:
            return (yield from cls._tile_reservoirs(op, in_df, weights))

    @classmethod
    def execute(cls, ctx, op: "DataFrameSample"):
        in_data = ctx[op.inputs[0].key]
        weights = op.weights
        if isinstance(weights, ENTITY_TYPE):
            weights = ctx[weights.key]

        size = op.size
        chunk_samples = op.chunk_samples
        if isinstance(chunk_samples, ENTITY_TYPE):
            chunk_samples = ctx[chunk_samples.key]
        if chunk_samples is not None:
            size = chunk_samples[op.inputs[0].index[op.axis]]

        ctx[op.outputs[0].key] = in_data.sample(
            n=size,
            frac=op.frac,
            replace=op.replace,
            weights=weights,
            random_state=op.random_state,
            axis=op.axis,
        )


def sample(
    df_or_series,
    n=None,
    frac=None,
    replace=False,
    weights=None,
    random_state=None,
    axis=None,
    always_multinomial=False,
):
    """
    Return a random sample of items from an axis of object.

    You can use `random_state` for reproducibility.

    Parameters
    ----------
    n : int, optional
        Number of items from axis to return. Cannot be used with `frac`.
        Default = 1 if `frac` = None.
    frac : float, optional
        Fraction of axis items to return. Cannot be used with `n`.
    replace : bool, default False
        Allow or disallow sampling of the same row more than once.
    weights : str or ndarray-like, optional
        Default 'None' results in equal probability weighting.
        If passed a Series, will align with target object on index. Index
        values in weights not found in sampled object will be ignored and
        index values in sampled object not in weights will be assigned
        weights of zero.
        If called on a DataFrame, will accept the name of a column
        when axis = 0.
        Unless weights are a Series, weights must be same length as axis
        being sampled.
        If weights do not sum to 1, they will be normalized to sum to 1.
        Missing values in the weights column will be treated as zero.
        Infinite values not allowed.
    random_state : int, array-like, BitGenerator, np.random.RandomState, optional
        If int, array-like, or BitGenerator (NumPy>=1.17), seed for
        random number generator
        If np.random.RandomState, use as numpy RandomState object.
    axis : {0 or ‘index’, 1 or ‘columns’, None}, default None
        Axis to sample. Accepts axis number or name. Default is stat axis
        for given data type (0 for Series and DataFrames).
    always_multinomial : bool, default False
        If True, always treat distribution of sample counts between data chunks
        as multinomial distribution. This will accelerate sampling when data
        is huge, but may affect randomness of samples when number of instances
        is not very large.

    Returns
    -------
    Series or DataFrame
        A new object of same type as caller containing `n` items randomly
        sampled from the caller object.

    See Also
    --------
    DataFrameGroupBy.sample: Generates random samples from each group of a
        DataFrame object.
    SeriesGroupBy.sample: Generates random samples from each group of a
        Series object.
    numpy.random.choice: Generates a random sample from a given 1-D numpy
        array.

    Notes
    -----
    If `frac` > 1, `replacement` should be set to `True`.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'num_legs': [2, 4, 8, 0],
    ...                    'num_wings': [2, 0, 0, 0],
    ...                    'num_specimen_seen': [10, 2, 1, 8]},
    ...                   index=['falcon', 'dog', 'spider', 'fish'])
    >>> df.execute()
            num_legs  num_wings  num_specimen_seen
    falcon         2          2                 10
    dog            4          0                  2
    spider         8          0                  1
    fish           0          0                  8

    Extract 3 random elements from the ``Series`` ``df['num_legs']``:
    Note that we use `random_state` to ensure the reproducibility of
    the examples.

    >>> df['num_legs'].sample(n=3, random_state=1).execute()
    fish      0
    spider    8
    falcon    2
    Name: num_legs, dtype: int64

    A random 50% sample of the ``DataFrame`` with replacement:

    >>> df.sample(frac=0.5, replace=True, random_state=1).execute()
          num_legs  num_wings  num_specimen_seen
    dog          4          0                  2
    fish         0          0                  8

    An upsample sample of the ``DataFrame`` with replacement:
    Note that `replace` parameter has to be `True` for `frac` parameter > 1.

    >>> df.sample(frac=2, replace=True, random_state=1).execute()
            num_legs  num_wings  num_specimen_seen
    dog            4          0                  2
    fish           0          0                  8
    falcon         2          2                 10
    falcon         2          2                 10
    fish           0          0                  8
    dog            4          0                  2
    fish           0          0                  8
    dog            4          0                  2

    Using a DataFrame column as weights. Rows with larger value in the
    `num_specimen_seen` column are more likely to be sampled.

    >>> df.sample(n=2, weights='num_specimen_seen', random_state=1).execute()
            num_legs  num_wings  num_specimen_seen
    falcon         2          2                 10
    fish           0          0                  8

    """
    axis = validate_axis(axis or 0, df_or_series)
    if axis == 1:
        raise NotImplementedError("Currently cannot sample over columns")
    rs = copy.deepcopy(
        random_state.to_numpy() if hasattr(random_state, "to_numpy") else random_state
    )
    op = DataFrameSample(
        size=n,
        frac=frac,
        replace=replace,
        weights=weights,
        random_state=rs,
        axis=axis,
        always_multinomial=always_multinomial,
    )
    return op(df_or_series)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy.special as spspecial

from ..arithmetic.utils import arithmetic_operand
from ..utils import infer_dtype, implement_scipy
from .core import (
    _register_special_op,
    TensorSpecialBinOp,
    TensorSpecialUnaryOp,
    TensorSpecialMultiOp,
)


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorEllipk(TensorSpecialUnaryOp):
    _func_name = "ellipk"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorEllipkm1(TensorSpecialUnaryOp):
    _func_name = "ellipkm1"


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorEllipkinc(TensorSpecialBinOp):
    _func_name = "ellipkinc"


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorEllipe(TensorSpecialUnaryOp):
    _func_name = "ellipe"


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorEllipeinc(TensorSpecialBinOp):
    _func_name = "ellipeinc"


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorElliprc(TensorSpecialBinOp):
    _func_name = "elliprc"


@_register_special_op
class TensorElliprd(TensorSpecialMultiOp):
    _ARG_COUNT = 3
    _func_name = "elliprd"


@_register_special_op
class TensorElliprf(TensorSpecialMultiOp):
    _ARG_COUNT = 3
    _func_name = "elliprf"


@_register_special_op
class TensorElliprg(TensorSpecialMultiOp):
    _ARG_COUNT = 3
    _func_name = "elliprg"


@_register_special_op
class TensorElliprj(TensorSpecialMultiOp):
    _ARG_COUNT = 4
    _func_name = "elliprj"


@implement_scipy(spspecial.ellipk)
@infer_dtype(spspecial.ellipk)
def ellipk(x, **kwargs):
    op = TensorEllipk(**kwargs)
    return op(x)


@implement_scipy(spspecial.ellipkm1)
@infer_dtype(spspecial.ellipkm1)
def ellipkm1(x, **kwargs):
    op = TensorEllipkm1(**kwargs)
    return op(x)


@implement_scipy(spspecial.ellipkinc)
@infer_dtype(spspecial.ellipkinc)
def ellipkinc(phi, m, **kwargs):
    op = TensorEllipkinc(**kwargs)
    return op(phi, m)


@implement_scipy(spspecial.ellipe)
@infer_dtype(spspecial.ellipe)
def ellipe(x, **kwargs):
    op = TensorEllipe(**kwargs)
    return op(x)


@implement_scipy(spspecial.ellipeinc)
@infer_dtype(spspecial.ellipeinc)
def ellipeinc(phi, m, **kwargs):
    op = TensorEllipeinc(**kwargs)
    return op(phi, m)


try:

    @implement_scipy(spspecial.elliprc)
    @infer_dtype(spspecial.elliprc)
    def elliprc(x, y, **kwargs):
        op = TensorElliprc(**kwargs)
        return op(x, y)

    @implement_scipy(spspecial.elliprd)
    @infer_dtype(spspecial.elliprd)
    def elliprd(x, y, z, **kwargs):
        op = TensorElliprd(**kwargs)
        return op(x, y, z)

    @implement_scipy(spspecial.elliprf)
    @infer_dtype(spspecial.elliprf)
    def elliprf(x, y, z, **kwargs):
        op = TensorElliprf(**kwargs)
        return op(x, y, z)

    @implement_scipy(spspecial.elliprg)
    @infer_dtype(spspecial.elliprg)
    def elliprg(x, y, z, **kwargs):
        op = TensorElliprg(**kwargs)
        return op(x, y, z)

    @implement_scipy(spspecial.elliprj)
    @infer_dtype(spspecial.elliprj)
    def elliprj(x, y, z, p, **kwargs):
        op = TensorElliprj(**kwargs)
        return op(x, y, z, p)

except AttributeError:
    # These functions are not implemented before scipy v1.8 so
    # spsecial.func may cause AttributeError
    elliprc = elliprd = elliprf = elliprg = elliprj = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="binary_or")
class TensorRshift(TensorBinOp):
    _op_type_ = OperandDef.RSHIFT
    _func_name = "right_shift"


@infer_dtype(np.right_shift)
def rshift(x1, x2, out=None, where=None, **kwargs):
    """
    Shift the bits of an integer to the right.

    Bits are shifted to the right `x2`.  Because the internal
    representation of numbers is in binary format, this operation is
    equivalent to dividing `x1` by ``2**x2``.

    Parameters
    ----------
    x1 : array_like, int
        Input values.
    x2 : array_like, int
        Number of bits to remove at the right of `x1`.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : Tensor, int
        Return `x1` with bits shifted `x2` times to the right.

    See Also
    --------
    left_shift : Shift the bits of an integer to the left.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> mt.right_shift(10, 1).execute()
    5

    >>> mt.right_shift(10, [1,2,3]).execute()
    array([5, 2, 1])
    """
    op = TensorRshift(**kwargs)
    return op(x1, x2, out=out, where=where)


@infer_dtype(np.right_shift, reverse=True)
def rrshift(x1, x2, **kwargs):
    op = TensorRshift(**kwargs)
    return op.rcall(x1, x2)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import warnings
import functools

logger = logging.getLogger(__name__)


# from https://github.com/numba/numba/blob/master/numba/core/entrypoints.py
# Must put this here to avoid extensions re-triggering initialization
@functools.lru_cache(maxsize=None)
def init_extension_entrypoints():
    """Execute all `mars_extensions` entry points with the name `init`
    If extensions have already been initialized, this function does nothing.
    """
    from pkg_resources import iter_entry_points

    for entry_point in iter_entry_points("mars_extensions", "init"):
        logger.info("Loading extension: %s", entry_point)
        try:
            func = entry_point.load()
            func()
        except Exception as e:
            msg = "Mars extension module '{}' failed to load due to '{}({})'."
            warnings.warn(
                msg.format(entry_point.module_name, type(e).__name__, str(e)),
                stacklevel=2,
            )
            logger.info("Extension loading failed for: %s", entry_point)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy.special as spspecial

from ..arithmetic.utils import arithmetic_operand
from ..utils import infer_dtype, implement_scipy
from .core import TensorSpecialBinOp, _register_special_op, TensorSpecialMultiOp


@_register_special_op
class TensorHYP2F1(TensorSpecialMultiOp):
    _ARG_COUNT = 4
    _func_name = "hyp2f1"


@implement_scipy(spspecial.hyp2f1)
@infer_dtype(spspecial.hyp2f1)
def hyp2f1(a, b, c, z, **kwargs):
    op = TensorHYP2F1(**kwargs)
    return op(a, b, c, z)


@_register_special_op
class TensorHYP1F1(TensorSpecialMultiOp):
    _ARG_COUNT = 3
    _func_name = "hyp1f1"


@implement_scipy(spspecial.hyp1f1)
@infer_dtype(spspecial.hyp1f1)
def hyp1f1(a, b, x, out=None, **kwargs):
    op = TensorHYP1F1(**kwargs)
    return op(a, b, x, out=out)


@_register_special_op
class TensorHYPERU(TensorSpecialMultiOp):
    _ARG_COUNT = 3
    _func_name = "hyperu"


@implement_scipy(spspecial.hyperu)
@infer_dtype(spspecial.hyperu)
def hyperu(a, b, x, out=None, **kwargs):
    op = TensorHYPERU(**kwargs)
    return op(a, b, x, out=out)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorHYP0F1(TensorSpecialBinOp):
    _func_name = "hyp0f1"


@implement_scipy(spspecial.hyp0f1)
@infer_dtype(spspecial.hyp0f1)
def hyp0f1(v, z, out=None, **kwargs):
    op = TensorHYP0F1(**kwargs)
    return op(v, z, out=out)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, List, Optional

from .... import oscar as mo
from ....utils import serialize_serializable, deserialize_serializable
from ...web import web_api, MarsServiceWebAPIHandler, MarsWebAPIClientMixin
from .core import AbstractMetaAPI


class MetaWebAPIHandler(MarsServiceWebAPIHandler):
    _root_pattern = "/api/session/(?P<session_id>[^/]+)/meta"

    async def _get_oscar_meta_api(self, session_id: str):
        from .oscar import MetaAPI

        return await self._get_api_by_key(MetaAPI, session_id)

    @web_api("(?P<data_key>[^/]+)", method="get")
    async def get_chunk_meta(self, session_id: str, data_key: str):
        fields_str = self.get_argument("fields", None)
        error = self.get_argument("error", "raise")
        fields = fields_str.split(",") if fields_str else None

        oscar_api = await self._get_oscar_meta_api(session_id)
        result = await oscar_api.get_chunk_meta(data_key, fields=fields, error=error)
        self.write(serialize_serializable(result))

    @web_api("", method="post")
    async def get_chunks_meta(self, session_id: str):
        body_args = deserialize_serializable(self.request.body)
        oscar_api = await self._get_oscar_meta_api(session_id)
        get_metas = []
        for data_key, fields, error in body_args:
            get_metas.append(oscar_api.get_chunk_meta.delay(data_key, fields, error))
        results = await oscar_api.get_chunk_meta.batch(*get_metas)
        self.write(serialize_serializable(results))


web_handlers = {MetaWebAPIHandler.get_root_pattern(): MetaWebAPIHandler}


class WebMetaAPI(AbstractMetaAPI, MarsWebAPIClientMixin):
    def __init__(
        self, session_id: str, address: str, request_rewriter: Callable = None
    ):
        # make sure all meta types registered
        from .. import metas

        del metas

        self._session_id = session_id
        self._address = address.rstrip("/")
        self.request_rewriter = request_rewriter

    @mo.extensible
    async def get_chunk_meta(
        self, object_id: str, fields: List[str] = None, error: str = "raise"
    ) -> Optional[Dict]:
        params = dict(error=error)
        req_addr = f"{self._address}/api/session/{self._session_id}/meta/{object_id}"
        if fields:
            params["fields"] = ",".join(fields)
        res = await self._request_url("GET", req_addr, params=params)
        return deserialize_serializable(res.body)

    @get_chunk_meta.batch
    async def get_chunks_meta(self, args_list, kwargs_list):
        get_chunk_metas = []
        for args, kwargs in zip(args_list, kwargs_list):
            object_id, fields, error = self.get_chunk_meta.bind(*args, **kwargs)
            get_chunk_metas.append([object_id, fields, error])

        req_addr = f"{self._address}/api/session/{self._session_id}/meta"
        res = await self._request_url(
            "POST", req_addr, data=serialize_serializable(get_chunk_metas)
        )
        return deserialize_serializable(res.body)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Iterator, Tuple
from ...core import ChunkGraph
from ...core.operand import (
    Fetch,
    FetchShuffle,
    MapReduceOperand,
    VirtualOperand,
)
from .core import Subtask


def iter_input_data_keys(
    subtask: Subtask,
    chunk_graph: ChunkGraph,
    chunk_key_to_data_keys: Dict[str, List[str]],
) -> Iterator[Tuple[str, bool]]:
    """An iterator yield (input data key, is shuffle)."""
    data_keys = set()
    for chunk in chunk_graph.iter_indep():
        if isinstance(chunk.op, Fetch) and chunk.key not in subtask.pure_depend_keys:
            data_keys.add(chunk.key)
            yield chunk.key, False
        elif isinstance(chunk.op, FetchShuffle):
            for key in chunk_key_to_data_keys[chunk.key]:
                if key not in data_keys:
                    data_keys.add(key)
                    yield key, True


def get_mapper_data_keys(key: str, context: Dict[str, Any]) -> List[str]:
    """Get the mapper data keys of key from context."""
    return [
        store_key
        for store_key in context
        if isinstance(store_key, tuple) and store_key[0] == key
    ]


def iter_output_data(
    chunk_graph: ChunkGraph, context: Dict[str, Any]
) -> Iterator[Tuple[str, Any, bool]]:
    """An iterator yield (output chunk key, output data, is shuffle)."""
    data_keys = set()
    for result_chunk in chunk_graph.result_chunks:
        # skip virtual operands for result chunks
        if isinstance(result_chunk.op, VirtualOperand):
            continue
        key = result_chunk.key
        if key in context:
            # non shuffle op
            data = context[key]
            # update meta
            if not isinstance(data, tuple):
                result_chunk.params = result_chunk.get_params_from_data(data)
            # check key after update meta
            if key in data_keys:
                continue
            yield key, data, False
            data_keys.add(key)
        else:
            assert isinstance(result_chunk.op, MapReduceOperand)
            keys = get_mapper_data_keys(key, context)
            for key in keys:
                if key in data_keys:
                    continue
                # shuffle op
                data = context[key]
                yield key, data, True
                data_keys.add(key)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np

from ... import execute
from ... import tensor as mt
from ...tensor.datasource import tensor as astensor


def softmax_loss_and_grad(W, X, y, reg):
    N, D = X.shape
    K = W.shape[1]

    y_obs = mt.zeros(shape=(N, K))
    for i in range(N):
        y_obs[i] = mt.eye(K)[y[i]]

    loss = -1 / N * mt.sum(
        y_obs * mt.log(mt.exp(X @ W) / mt.sum(mt.exp(X @ W), axis=1).reshape(-1, 1))
    ) + 0.5 * reg * mt.sum(mt.square(W))

    dW = mt.zeros(shape=(D, K))

    # Matrix approach
    dW = (
        -1
        / N
        * X.T
        @ (y_obs - (mt.exp(X @ W) / mt.sum(mt.exp(X @ W), axis=1).reshape(-1, 1)))
        + reg * W
    )

    execute(loss, dW)

    return loss, dW


def gradient_descent(
    X,
    y,
    learning_rate=1e-3,
    reg=1e-5,
    max_epochs=100,
    batch_size=20,
    fit_intercept=True,
    verbose=0,
):
    # assume y takes values 0...K-1 where K is number of classes
    num_classes = (mt.max(y) + 1).to_numpy()

    num_train, dim = X.shape
    num_iters_per_epoch = int(math.floor(1.0 * num_train / batch_size))

    # need extra entries if fit_intercept
    if fit_intercept:
        X = mt.hstack((X, mt.ones((num_train, 1))))
        W = 0.001 * mt.random.randn(dim + 1, num_classes).execute()
    else:
        X = astensor(X)
        W = 0.001 * mt.random.randn(dim, num_classes).execute()

    for _ in range(max_epochs):
        # perform mini-batch SGD update
        perm_idx = np.random.permutation(num_train)
        for it in range(num_iters_per_epoch):
            # print(it, num_iters_per_epoch)
            idx = perm_idx[it * batch_size : (it + 1) * batch_size]
            batch_x = X[idx]
            batch_y = y[idx]

            # evaluate loss and gradient
            _, grad = softmax_loss_and_grad(W, batch_x, batch_y, reg)

            # update parameters
            W = W - learning_rate * grad

    return W



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core import ChunkType
from ...tensor import arithmetic
from ...tensor import reduction
from ...tensor.fuse import TensorNeFuseChunk
from ...tensor.fuse.numexpr import NUMEXPR_INSTALLED
from .core import RuntimeOptimizer, register_optimizer


REDUCTION_OP = {
    reduction.TensorSum,
    reduction.TensorProd,
    reduction.TensorMax,
    reduction.TensorMin,
}
SUPPORT_OP = {
    arithmetic.TensorAdd,
    arithmetic.TensorSubtract,
    arithmetic.TensorMultiply,
    arithmetic.TensorDivide,
    arithmetic.TensorPower,
    arithmetic.TensorMod,
    arithmetic.TensorNegative,
    arithmetic.TensorAbs,
    arithmetic.TensorConj,
    arithmetic.TensorExp,
    arithmetic.TensorLog,
    arithmetic.TensorLog10,
    arithmetic.TensorExpm1,
    arithmetic.TensorLog1p,
    arithmetic.TensorSqrt,
    arithmetic.TensorEqual,
    arithmetic.TensorNotEqual,
    arithmetic.TensorLessThan,
    arithmetic.TensorLessEqual,
    arithmetic.TensorGreaterThan,
    arithmetic.TensorGreaterEqual,
    arithmetic.TensorSin,
    arithmetic.TensorCos,
    arithmetic.TensorTan,
    arithmetic.TensorArcsin,
    arithmetic.TensorArccos,
    arithmetic.TensorArctan,
    arithmetic.TensorSinh,
    arithmetic.TensorCosh,
    arithmetic.TensorTanh,
    arithmetic.TensorArcsinh,
    arithmetic.TensorArccosh,
    arithmetic.TensorArctanh,
    arithmetic.TensorLshift,
    arithmetic.TensorRshift,
    arithmetic.TensorTreeAdd,
    arithmetic.TensorTreeMultiply,
    reduction.TensorSum,
    reduction.TensorProd,
    reduction.TensorMax,
    reduction.TensorMin,
}


def _check_reduction_axis(node: ChunkType):
    return len(node.op.axis) == 1 or len(node.op.axis) == node.ndim


def _support(node: ChunkType):
    op_type = type(node.op)
    if op_type in REDUCTION_OP:
        return _check_reduction_axis(node)
    return op_type in SUPPORT_OP


def _transfer_op(node: ChunkType):
    op = node.op
    if type(op) in REDUCTION_OP and not _check_reduction_axis(node):
        return op
    return op


@register_optimizer
class NumexprRuntimeOptimizer(RuntimeOptimizer):
    engine = "numexpr"

    @classmethod
    def is_available(cls) -> bool:
        return NUMEXPR_INSTALLED

    def optimize(self):
        fuses = []
        explored = set()

        graph = self._graph
        for node in graph.topological_iter():
            if node.op.gpu or node.op.sparse:
                # break
                return [], []
            if type(node.op) not in SUPPORT_OP or node in graph.results:
                continue
            if node in explored or type(node.op) in REDUCTION_OP:
                # TODO: check logic here
                continue
            if graph.count_successors(node) != 1:
                continue

            selected = [node]
            # add successors
            cur_node = graph.successors(node)[0]
            while graph.count_predecessors(cur_node) == 1 and _support(cur_node):
                selected.append(cur_node)
                if (
                    graph.count_successors(cur_node) != 1
                    or type(cur_node.op) in REDUCTION_OP
                    or cur_node in graph.results
                ):
                    break
                else:
                    cur_node = graph.successors(cur_node)[0]
            if len(selected) > 1:
                explored.update(selected)
                fuses.append(list(selected))

        return self._fuse_nodes(fuses, TensorNeFuseChunk)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures as futures
import contextlib
import logging.config
import multiprocessing
import os
import random
import signal
import sys
import threading
import uuid
from dataclasses import dataclass
from types import TracebackType
from typing import List

from ....utils import dataslots, ensure_coverage, reset_id_random_seed
from ..config import ActorPoolConfig
from ..message import CreateActorMessage
from ..pool import MainActorPoolBase, SubActorPoolBase, _register_message_handler


_is_windows: bool = sys.platform.startswith("win")

if sys.version_info[:2] == (3, 9):
    # fix for Python 3.9, see https://bugs.python.org/issue43517
    if sys.platform == "win32":
        from multiprocessing import popen_spawn_win32 as popen_spawn

        popen_forkserver = popen_fork = synchronize = None
    else:
        from multiprocessing import popen_spawn_posix as popen_spawn
        from multiprocessing import popen_forkserver, popen_fork, synchronize
    _ = popen_spawn, popen_forkserver, popen_fork, synchronize
    del _
elif sys.version_info[:2] == (3, 6):  # pragma: no cover
    from multiprocessing.process import BaseProcess

    # define kill method for multiprocessing
    def _mp_kill(self):
        if not _is_windows:
            try:
                os.kill(self.pid, signal.SIGKILL)
            except ProcessLookupError:
                pass
            except OSError:
                if self.wait(timeout=0.1) is None:
                    raise
        else:
            self.terminate()

    BaseProcess.kill = _mp_kill

logger = logging.getLogger(__name__)
_init_main_suspended_local = threading.local()


def _patch_spawn_get_preparation_data():
    try:
        from multiprocessing import spawn as mp_spawn

        _raw_get_preparation_data = mp_spawn.get_preparation_data

        def _patched_get_preparation_data(*args, **kw):
            ret = _raw_get_preparation_data(*args, **kw)
            if getattr(_init_main_suspended_local, "value", False):
                # make sure user module is not imported when start Mars cluster
                ret.pop("init_main_from_name", None)
                ret.pop("init_main_from_path", None)
            return ret

        _patched_get_preparation_data._mars_patched = True
        if not getattr(mp_spawn.get_preparation_data, "_mars_patched", False):
            mp_spawn.get_preparation_data = _patched_get_preparation_data
    except (ImportError, AttributeError):  # pragma: no cover
        pass


@contextlib.contextmanager
def _suspend_init_main():
    try:
        _init_main_suspended_local.value = True
        yield
    finally:
        _init_main_suspended_local.value = False


@dataslots
@dataclass
class SubpoolStatus:
    # for status, 0 is succeeded, 1 is failed
    status: int = None
    external_addresses: List[str] = None
    error: BaseException = None
    traceback: TracebackType = None


@_register_message_handler
class MainActorPool(MainActorPoolBase):
    @classmethod
    def get_external_addresses(
        cls, address: str, n_process: int = None, ports: List[int] = None
    ):
        """Get socket address for every process"""
        if ":" in address:
            host, port = address.split(":", 1)
            port = int(port)
            if ports:
                if len(ports) != n_process:
                    raise ValueError(
                        f"`ports` specified, but its count "
                        f"is not equal to `n_process`, "
                        f"number of ports: {len(ports)}, "
                        f"n_process: {n_process}"
                    )
                sub_ports = ports
            else:
                sub_ports = [0] * n_process
        else:
            host = address
            if ports and len(ports) != n_process + 1:
                # ports specified, the first of which should be main port
                raise ValueError(
                    f"`ports` specified, but its count "
                    f"is not equal to `n_process` + 1, "
                    f"number of ports: {len(ports)}, "
                    f"n_process + 1: {n_process + 1}"
                )
            elif not ports:
                ports = [0] * (n_process + 1)
            port = ports[0]
            sub_ports = ports[1:]
        return [f"{host}:{port}" for port in [port] + sub_ports]

    @classmethod
    def gen_internal_address(
        cls, process_index: int, external_address: str = None
    ) -> str:
        if hasattr(asyncio, "start_unix_server"):
            return f"unixsocket:///{process_index}"
        else:
            return external_address

    @classmethod
    async def start_sub_pool(
        cls,
        actor_pool_config: ActorPoolConfig,
        process_index: int,
        start_method: str = None,
    ):
        def start_pool_in_process():
            ctx = multiprocessing.get_context(method=start_method)
            status_queue = ctx.Queue()

            with _suspend_init_main():
                process = ctx.Process(
                    target=cls._start_sub_pool,
                    args=(actor_pool_config, process_index, status_queue),
                    name=f"MarsActorPool{process_index}",
                )
                process.daemon = True
                process.start()

            # wait for sub actor pool to finish starting
            process_status = status_queue.get()
            return process, process_status

        _patch_spawn_get_preparation_data()
        loop = asyncio.get_running_loop()
        with futures.ThreadPoolExecutor(1) as executor:
            create_pool_task = loop.run_in_executor(executor, start_pool_in_process)
            return await create_pool_task

    @classmethod
    async def wait_sub_pools_ready(cls, create_pool_tasks: List[asyncio.Task]):
        processes = []
        ext_addresses = []
        for task in create_pool_tasks:
            process, status = await task
            if status.status == 1:
                # start sub pool failed
                raise status.error.with_traceback(status.traceback)
            processes.append(process)
            ext_addresses.append(status.external_addresses)
        return processes, ext_addresses

    @classmethod
    def _start_sub_pool(
        cls,
        actor_config: ActorPoolConfig,
        process_index: int,
        status_queue: multiprocessing.Queue,
    ):
        ensure_coverage()

        # make sure enough randomness for every sub pool
        random.seed(uuid.uuid1().bytes)
        reset_id_random_seed()

        conf = actor_config.get_pool_config(process_index)
        suspend_sigint = conf["suspend_sigint"]
        if suspend_sigint:
            signal.signal(signal.SIGINT, lambda *_: None)

        logging_conf = conf["logging_conf"] or {}
        if logging_conf.get("file"):
            logging.config.fileConfig(logging_conf["file"])
        elif logging_conf.get("level"):
            logging.getLogger("__main__").setLevel(logging_conf["level"])
            logging.getLogger("mars").setLevel(logging_conf["level"])
            if logging_conf.get("format"):
                logging.basicConfig(format=logging_conf["format"])

        use_uvloop = conf["use_uvloop"]
        if use_uvloop:
            import uvloop

            asyncio.set_event_loop(uvloop.new_event_loop())
        else:
            asyncio.set_event_loop(asyncio.new_event_loop())

        coro = cls._create_sub_pool(actor_config, process_index, status_queue)
        asyncio.run(coro)

    @classmethod
    async def _create_sub_pool(
        cls,
        actor_config: ActorPoolConfig,
        process_index: int,
        status_queue: multiprocessing.Queue,
    ):
        process_status = None
        try:
            cur_pool_config = actor_config.get_pool_config(process_index)
            env = cur_pool_config["env"]
            if env:
                os.environ.update(env)
            pool = await SubActorPool.create(
                {"actor_pool_config": actor_config, "process_index": process_index}
            )
            external_addresses = cur_pool_config["external_address"]
            process_status = SubpoolStatus(
                status=0, external_addresses=external_addresses
            )
            await pool.start()
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except
            _, error, tb = sys.exc_info()
            process_status = SubpoolStatus(status=1, error=error, traceback=tb)
            raise
        finally:
            status_queue.put(process_status)
        await pool.join()

    async def kill_sub_pool(
        self, process: multiprocessing.Process, force: bool = False
    ):
        if (
            "COV_CORE_SOURCE" in os.environ and not force and not _is_windows
        ):  # pragma: no cover
            # must shutdown gracefully, or coverage info lost
            try:
                os.kill(process.pid, signal.SIGINT)
            except OSError:  # pragma: no cover
                pass
            process.terminate()
            wait_pool = futures.ThreadPoolExecutor(1)
            try:
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(wait_pool, process.join, 3)
            finally:
                wait_pool.shutdown(False)
        process.kill()
        await asyncio.to_thread(process.join, 5)

    async def is_sub_pool_alive(self, process: multiprocessing.Process):
        try:
            return await asyncio.to_thread(process.is_alive)
        except RuntimeError as ex:  # pragma: no cover
            if "shutdown" not in str(ex):
                # when atexit is triggered, the default pool might be shutdown
                # and to_thread will fail
                raise
            return process.is_alive()

    async def recover_sub_pool(self, address: str):
        process_index = self._config.get_process_index(address)
        # process dead, restart it
        # remember always use spawn to recover sub pool
        task = asyncio.create_task(
            self.start_sub_pool(self._config, process_index, "spawn")
        )
        self.sub_processes[address] = (await self.wait_sub_pools_ready([task]))[0][0]

        if self._auto_recover == "actor":
            # need to recover all created actors
            for _, message in self._allocated_actors[address].values():
                create_actor_message: CreateActorMessage = message
                await self.call(address, create_actor_message)

    async def start(self):
        await super().start()
        await self.start_monitor()


@_register_message_handler
class SubActorPool(SubActorPoolBase):
    pass
